"""
Database setup
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

import logging
from hub.persistence.repository import Repository
from hub.persistence.models import Application
from hub.persistence.models import City
from hub.persistence.models import CityObject
from hub.persistence.models import User
from hub.persistence.models import UserRoles
from hub.persistence.models import SimulationResults
from hub.persistence.repositories.user import User as UserRepository
from hub.persistence.repositories.application import Application as ApplicationRepository


class DBSetup:
  """
  Creates a Persistence database structure
  """

  def __init__(self, db_name, app_env, dotenv_path, admin_password, application_uuid):
    """
    Creates database tables a default admin user and a default admin app with the given password and uuid
    :param db_name: database name
    :param app_env: application environment type [TEST|PROD]
    :param dotenv_path: .env file path
    :param admin_password: administrator password for the application uuid
    :application_uuid: application uuid
    """
    repository = Repository(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)

    # Create the tables using the models
    Application.__table__.create(bind=repository.engine, checkfirst=True)
    User.__table__.create(bind=repository.engine, checkfirst=True)
    City.__table__.create(bind=repository.engine, checkfirst=True)
    CityObject.__table__.create(bind=repository.engine, checkfirst=True)
    SimulationResults.__table__.create(bind=repository.engine, checkfirst=True)

    self._user_repo = UserRepository(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    self._application_repo = ApplicationRepository(db_name=db_name, app_env=app_env, dotenv_path=dotenv_path)
    application_id = self._create_admin_app(self._application_repo, application_uuid)
    self._create_admin_user(self._user_repo, admin_password, application_id)

  @staticmethod
  def _create_admin_app(application_repo, application_uuid):
    name = 'AdminTool'
    description = 'Admin tool to control city persistence and to test the API v1.4'
    logging.info('Creating default admin tool application...')
    application = application_repo.insert(name, description, application_uuid)

    if isinstance(application, dict):
      logging.info(application)
    else:
      msg = f'Created Admin tool with application_uuid: {application_uuid}'
      logging.info(msg)
    return application.id

  @staticmethod
  def _create_admin_user(user_repo, admin_password, application_id):
    password = admin_password
    logging.info('Creating default admin user...')
    user = user_repo.insert('Administrator', password, UserRoles.Admin, application_id)
    if isinstance(user, dict):
      logging.info(user)
    else:
      logging.info('Created Admin user')



"""
heat_pump module defines a heat pump
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
from typing import List
from pandas.core.series import Series


class HeatPump:
  """
  HeatPump class
  """

  def __init__(self):
    self._model = None
    self._hp_monthly_fossil_consumption = None
    self._hp_monthly_electricity_demand = None

  @property
  def model(self) -> str:
    """
    Get model name
    :return: str
    """
    return self._model

  @model.setter
  def model(self, value):
    """
    Set model (name, indicated in capacity)
    :param value: str
    """
    if self._model is None:
      self._model = value

  @property
  def hp_monthly_fossil_consumption(self) -> List:
    """
    Fossil fuel consumption that results from insel simulation
    ":return: []
    :return:
    """
    return self._hp_monthly_fossil_consumption

  @hp_monthly_fossil_consumption.setter
  def hp_monthly_fossil_consumption(self, value):
    if isinstance(value, Series):
      self._hp_monthly_fossil_consumption = value

  @property
  def hp_monthly_electricity_demand(self) -> List:
    """
    Electricity demand that results from insel simulation
    ":return: []
    :return:
    """
    return self._hp_monthly_electricity_demand

  @hp_monthly_electricity_demand.setter
  def hp_monthly_electricity_demand(self, value):
    if isinstance(value, Series):
      self._hp_monthly_electricity_demand = value



"""
Sensor module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.helpers.location import Location
from hub.city_model_structure.iot.sensor_measure import SensorMeasure
from hub.city_model_structure.iot.sensor_type import SensorType


class Sensor:
  """
  Sensor abstract class
  """
  def __init__(self):
    self._name = None
    self._type = None
    self._units = None
    self._location = None

  @property
  def name(self):
    """
    Get sensor name
    :return: str
    """
    if self._name is None:
      raise ValueError('Undefined sensor name')
    return self._name

  @name.setter
  def name(self, value):
    """
    Set sensor name
    :param value: str
    """
    if value is not None:
      self._name = str(value)

  @property
  def type(self) -> SensorType:
    """
    Get sensor type
    :return: str
    """
    return self._type

  @property
  def units(self):
    """
    Get sensor units
    :return: str
    """
    return self._units

  @property
  def location(self) -> Location:
    """
    Get sensor location
    :return: Location
    """
    return self._location

  @location.setter
  def location(self, value):
    """
    Set sensor location
    :param value: Location
    """
    self._location = value

  @property
  def measures(self) -> [SensorMeasure]:
    """
    Raises not implemented error
    """
    raise NotImplementedError



"""
Simulation results repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""
import datetime
import logging

from sqlalchemy import or_
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from hub.persistence.repository import Repository
from hub.persistence.models import City
from hub.persistence.models import CityObject
from hub.persistence.models import SimulationResults as Model


class SimulationResults(Repository):
  """
  Simulation results repository
  """
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(SimulationResults, cls).__new__(cls)
    return cls._instance

  def insert(self, name: str, values: str, city_id=None, city_object_id=None):
    """
    Inserts simulations results linked either with a city as a whole or with a city object
    :param name: results name
    :param values: the simulation results in json format
    :param city_id: optional city id
    :param city_object_id: optional city object id
    :return: Identity id
    """
    if city_id is not None:
      _ = self._get_city(city_id)
    else:
      _ = self._get_city_object(city_object_id)
    try:
      simulation_result = Model(name=name,
                                values=values,
                                city_id=city_id,
                                city_object_id=city_object_id)
      self.session.add(simulation_result)
      self.session.flush()
      self.session.commit()
      self.session.refresh(simulation_result)
      return simulation_result.id
    except SQLAlchemyError as err:
      logging.error('An error occurred while creating city_object %s', err)
      raise SQLAlchemyError from err

  def update(self, name: str, values: str, city_id=None, city_object_id=None):
    """
    Updates simulation results for a city or a city object
    :param name: The simulation results tool and workflow name
    :param values: the simulation results in json format
    :param city_id: optional city id
    :param city_object_id: optional city object id
    :return: None
    """
    try:
      if city_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_id == city_id).update(
          {
            'values': values,
            'updated': datetime.datetime.utcnow()
          })
        self.session.commit()
      elif city_object_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_object_id == city_object_id).update(
          {
            'values': values,
            'updated': datetime.datetime.utcnow()
          })
        self.session.commit()
      else:
        raise NotImplementedError('Missing either city_id or city_object_id')
    except SQLAlchemyError as err:
      logging.error('Error while updating city object %s', err)
      raise SQLAlchemyError from err

  def delete(self, name: str, city_id=None, city_object_id=None):
    """
    Deletes an application with the application_uuid
    :param name: The simulation results tool and workflow name
    :param city_id: The id for the city owning the simulation results
    :param city_object_id: the id for the city_object owning these simulation results
    :return: None
    """
    try:
      if city_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_id == city_id).delete()
        self.session.commit()
      elif city_object_id is not None:
        self.session.query(Model).filter(Model.name == name, Model.city_object_id == city_object_id).delete()
        self.session.commit()
      else:
        raise NotImplementedError('Missing either city_id or city_object_id')
    except SQLAlchemyError as err:
      logging.error('Error while deleting application: %s', err)
      raise SQLAlchemyError from err

  def _get_city(self, city_id) -> City:
    """
    Fetch a city object based city id
    :param city_id: a city identifier
    :return: [City] with the provided city_id
    """
    try:
      return self.session.execute(select(City).where(City.id == city_id)).first()
    except SQLAlchemyError as err:
      logging.error('Error while fetching city by city_id: %s', err)
      raise SQLAlchemyError from err

  def _get_city_object(self, city_object_id) -> [CityObject]:
    """
    Fetch a city object based city id
    :param city_object_id: a city object identifier
    :return: [CityObject] with the provided city_object_id
    """
    try:
      return self.session.execute(select(CityObject).where(CityObject.id == city_object_id)).first()
    except SQLAlchemyError as err:
      logging.error('Error while fetching city by city_id: %s', err)
      raise SQLAlchemyError from err

  def get_simulation_results_by_city_id_city_object_id_and_names(self, city_id, city_object_id, result_names=None) -> [Model]:
    """
    Fetch the simulation results based in the city_id or city_object_id with the given names or all
    :param city_id: the city id
    :param city_object_id: the city object id
    :param result_names: if given filter the results
    :return: [SimulationResult]
    """
    try:
      result_set = self.session.execute(select(Model).where(or_(
        Model.city_id == city_id,
        Model.city_object_id == city_object_id
      )))
      results = [r[0] for r in result_set]
      if not result_names:
        return results
      filtered_results = []
      for result in results:
        if result.name in result_names:
          filtered_results.append(result)
      return filtered_results
    except SQLAlchemyError as err:
      logging.error('Error while fetching city by city_id: %s', err)
      raise SQLAlchemyError from err



"""
Greenery catalog data model Plant percentage class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.catalog_factories.data_models.greenery.plant import Plant as HubPlant


class PlantPercentage(HubPlant):
  """
  Plant percentage class
  """

  def __init__(self, percentage, plant_category, plant):
    super().__init__(plant_category, plant)
    self._percentage = percentage

  @property
  def percentage(self):
    """
    Get plant percentage
    :return: float
    """
    return self._percentage

  def to_dictionary(self):
    """Class content to dictionary"""
    _soils = []
    for _soil in self.grows_on:
      _soils.append(_soil.to_dictionary())
    content = {'Plant': {'name': self.name,
                         'percentage': self.percentage,
                         'category': self.category,
                         'height [m]': self.height,
                         'leaf area index': self.leaf_area_index,
                         'leaf reflectivity': self.leaf_reflectivity,
                         'leaf emissivity': self.leaf_emissivity,
                         'minimal stomatal resistance [s/m]': self.minimal_stomatal_resistance,
                         'co2 sequestration [kg????]': self.co2_sequestration,
                         'soils where it grows on': _soils
                         }
               }

    return content



"""
Greenery catalog content
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class Content:
  """
  Content class
  """
  def __init__(self, vegetations, plants, soils):
    self._vegetations = vegetations
    self._plants = plants
    self._soils = soils

  @property
  def vegetations(self):
    """
    All vegetation in the catalog
    """
    return self._vegetations

  @property
  def plants(self):
    """
    All plants in the catalog
    """
    return self._plants

  @property
  def soils(self):
    """
    All soils in the catalog
    """
    return self._soils

  def to_dictionary(self):
    """Class content to dictionary"""
    _archetypes = []
    for _archetype in self.vegetations:
      _archetypes.append(_archetype.to_dictionary())
    content = {'Archetypes': _archetypes}

    return content

  def __str__(self):
    """Print content"""
    _archetypes = []
    for _archetype in self.vegetations:
      _archetypes.append(_archetype.to_dictionary())
    content = {'Archetypes': _archetypes}

    return str(content)



"""
Models package
"""
from .application import Application
from .city import City
from .city_object import CityObject
from .simulation_results import SimulationResults
from .user import User, UserRoles



"""
Building module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging
from typing import List, Union, TypeVar

import numpy as np
import hub.helpers.constants as cte
from hub.city_model_structure.attributes.polyhedron import Polyhedron
from hub.city_model_structure.building_demand.household import Household
from hub.city_model_structure.building_demand.internal_zone import InternalZone
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.city_object import CityObject
from hub.city_model_structure.energy_systems.energy_system import EnergySystem
from hub.helpers.peak_loads import PeakLoads

City = TypeVar('City')


class Building(CityObject):
  """
  Building(CityObject) class
  """
  def __init__(self, name, surfaces, year_of_construction, function, terrains=None, city=None):
    super().__init__(name, surfaces)
    self._city = city
    self._households = None
    self._basement_heated = None
    self._attic_heated = None
    self._terrains = terrains
    self._year_of_construction = year_of_construction
    self._function = function
    self._average_storey_height = None
    self._storeys_above_ground = None
    self._floor_area = None
    self._roof_type = None
    self._internal_zones = None
    self._thermal_zones_from_internal_zones = None
    self._shell = None
    self._aliases = []
    self._type = 'building'
    self._cold_water_temperature = {}
    self._heating_demand = {}
    self._cooling_demand = {}
    self._lighting_electrical_demand = {}
    self._appliances_electrical_demand = {}
    self._domestic_hot_water_heat_demand = {}
    self._heating_consumption = {}
    self._cooling_consumption = {}
    self._domestic_hot_water_consumption = {}
    self._distribution_systems_electrical_consumption = {}
    self._onsite_electrical_production = {}
    self._eave_height = None
    self._energy_systems = None
    self._systems_archetype_name = None
    self._grounds = []
    self._roofs = []
    self._walls = []
    self._internal_walls = []
    self._ground_walls = []
    self._attic_floors = []
    self._interior_slabs = []
    for surface_id, surface in enumerate(self.surfaces):
      self._min_x = min(self._min_x, surface.lower_corner[0])
      self._min_y = min(self._min_y, surface.lower_corner[1])
      self._min_z = min(self._min_z, surface.lower_corner[2])
      surface.id = surface_id
      if surface.type == cte.GROUND:
        self._grounds.append(surface)
      elif surface.type == cte.WALL:
        self._walls.append(surface)
      elif surface.type == cte.ROOF:
        self._roofs.append(surface)
      elif surface.type == cte.INTERIOR_WALL:
        self._internal_walls.append(surface)
      elif surface.type == cte.GROUND_WALL:
        self._ground_walls.append(surface)
      elif surface.type == cte.ATTIC_FLOOR:
        self._attic_floors.append(surface)
      elif surface.type == cte.INTERIOR_SLAB:
        self._interior_slabs.append(surface)
      else:
        logging.error(f'Building %s [%s] has an unexpected surface type %s.', self.name, self.aliases, surface.type)

  @property
  def shell(self) -> Polyhedron:
    """
    Get building's external polyhedron
    :return: [Polyhedron]
    """
    polygons = []
    for surface in self.surfaces:
      if surface.type is not cte.INTERIOR_WALL:
        polygons.append(surface.solid_polygon)
        if surface.holes_polygons is not None:
          for hole in surface.holes_polygons:
            polygons.append(hole)
    if self._shell is None:
      self._shell = Polyhedron(polygons)
    return self._shell

  @property
  def internal_zones(self) -> List[InternalZone]:
    """
    Get building internal zones
    For Lod up to 3, there is only one internal zone which corresponds to the building shell.
    In LoD 4 there can be more than one. In this case the definition of surfaces and floor area must be redefined.
    :return: [InternalZone]
    """
    if self._internal_zones is None:
      self._internal_zones = [InternalZone(self.surfaces, self.floor_area, self.volume)]
    return self._internal_zones

  @property
  def thermal_zones_from_internal_zones(self) -> Union[None, List[ThermalZone]]:
    """
    Get building thermal zones
    :return: [ThermalZone]
    """
    if self._thermal_zones_from_internal_zones is None:
      self._thermal_zones_from_internal_zones = []
      for internal_zone in self.internal_zones:
        if internal_zone.thermal_zones_from_internal_zones is None:
          self._thermal_zones_from_internal_zones = None
          return self._thermal_zones_from_internal_zones
        self._thermal_zones_from_internal_zones.append(internal_zone.thermal_zones_from_internal_zones[0])
    return self._thermal_zones_from_internal_zones

  @property
  def grounds(self) -> List[Surface]:
    """
    Get building ground surfaces
    :return: [Surface]
    """
    return self._grounds

  @property
  def roofs(self) -> List[Surface]:
    """
    Get building roof surfaces
    :return: [Surface]
    """
    return self._roofs

  @property
  def walls(self) -> List[Surface]:
    """
    Get building wall surfaces
    :return: [Surface]
    """
    return self._walls

  @property
  def internal_walls(self) -> List[Surface]:
    """
    Get building internal wall surfaces
    :return: [Surface]
    """
    return self._internal_walls

  @property
  def terrains(self) -> Union[None, List[Surface]]:
    """
    Get city object terrain surfaces
    :return: [Surface]
    """
    return self._terrains

  @property
  def attic_heated(self) -> Union[None, int]:
    """
    Get if the city object attic is heated
    0: no attic in the building
    1: attic exists but is not heated
    2: attic exists and is heated
    :return: None or int
    """
    return self._attic_heated

  @attic_heated.setter
  def attic_heated(self, value):
    """
    Set if the city object attic is heated
    0: no attic in the building
    1: attic exists but is not heated
    2: attic exists and is heated
    :param value: int
    """
    if value is not None:
      self._attic_heated = int(value)

  @property
  def basement_heated(self) -> Union[None, int]:
    """
    Get if the city object basement is heated
    0: no basement in the building
    1: basement exists but is not heated
    2: basement exists and is heated
    :return: None or int
    """
    return self._basement_heated

  @basement_heated.setter
  def basement_heated(self, value):
    """
    Set if the city object basement is heated
    0: no basement in the building
    1: basement exists but is not heated
    2: basement exists and is heated
    :param value: int
    """
    if value is not None:
      self._basement_heated = int(value)

  @property
  def year_of_construction(self):
    """
    Get building year of construction
    :return: int
    """
    return self._year_of_construction

  @year_of_construction.setter
  def year_of_construction(self, value):
    """
    Set building year of construction
    :param value: int
    """
    if value is not None:
      self._year_of_construction = int(value)

  @property
  def function(self) -> Union[None, str]:
    """
    Get building function
    :return: None or str
    """
    return self._function

  @function.setter
  def function(self, value):
    """
    Set building function
    :param value: str
    """
    if value is not None:
      self._function = str(value)

  @property
  def average_storey_height(self) -> Union[None, float]:
    """
    Get building average storey height in meters
    :return: None or float
    """
    if len(self.internal_zones) > 1:
      self._average_storey_height = 0
      for internal_zone in self.internal_zones:
        self._average_storey_height += internal_zone.mean_height / len(self.internal_zones)
    else:
      if self.internal_zones[0].thermal_archetype is None:
        self._average_storey_height = None
      else:
        self._average_storey_height = self.internal_zones[0].thermal_archetype.average_storey_height
    return self._average_storey_height

  @average_storey_height.setter
  def average_storey_height(self, value):
    """
    Set building average storey height in meters
    :param value: float
    """
    if value is not None:
      self._average_storey_height = float(value)

  @property
  def storeys_above_ground(self) -> Union[None, int]:
    """
    Get building storeys number above ground
    :return: None or int
    """
    if self._storeys_above_ground is None:
      if self.eave_height is not None and self.average_storey_height is not None:
        self._storeys_above_ground = int(self.eave_height / self.average_storey_height)
    return self._storeys_above_ground

  @storeys_above_ground.setter
  def storeys_above_ground(self, value):
    """
    Set building storeys number above ground
    :param value: int
    """
    if value is not None:
      self._storeys_above_ground = int(value)

  @property
  def cold_water_temperature(self) -> {float}:
    """
    Get cold water temperature in degrees Celsius
    :return: dict{[float]}
    """
    return self._cold_water_temperature

  @cold_water_temperature.setter
  def cold_water_temperature(self, value):
    """
    Set cold water temperature in degrees Celsius
    :param value: dict{[float]}
    """
    self._cold_water_temperature = value

  @property
  def heating_demand(self) -> dict:
    """
    Get heating demand in J
    :return: dict{[float]}
    """
    return self._heating_demand

  @heating_demand.setter
  def heating_demand(self, value):
    """
    Set heating demand in J
    :param value: dict{[float]}
    """
    self._heating_demand = value

  @property
  def cooling_demand(self) -> dict:
    """
    Get cooling demand in J
    :return: dict{[float]}
    """
    return self._cooling_demand

  @cooling_demand.setter
  def cooling_demand(self, value):
    """
    Set cooling demand in J
    :param value: dict{[float]}
    """
    self._cooling_demand = value

  @property
  def lighting_electrical_demand(self) -> dict:
    """
    Get lighting electrical demand in J
    :return: dict{[float]}
    """
    return self._lighting_electrical_demand

  @lighting_electrical_demand.setter
  def lighting_electrical_demand(self, value):
    """
    Set lighting electrical demand in J
    :param value: dict{[float]}
    """
    self._lighting_electrical_demand = value

  @property
  def appliances_electrical_demand(self) -> dict:
    """
    Get appliances electrical demand in J
    :return: dict{[float]}
    """
    return self._appliances_electrical_demand

  @appliances_electrical_demand.setter
  def appliances_electrical_demand(self, value):
    """
    Set appliances electrical demand in J
    :param value: dict{[float]}
    """
    self._appliances_electrical_demand = value

  @property
  def domestic_hot_water_heat_demand(self) -> dict:
    """
    Get domestic hot water heat demand in J
    :return: dict{[float]}
    """
    return self._domestic_hot_water_heat_demand

  @domestic_hot_water_heat_demand.setter
  def domestic_hot_water_heat_demand(self, value):
    """
    Set domestic hot water heat demand in J
    :param value: dict{[float]}
    """
    self._domestic_hot_water_heat_demand = value

  @property
  def lighting_peak_load(self) -> Union[None, dict]:
    """
    Get lighting peak load in W
    :return: dict{[float]}
    """
    results = {}
    peak_lighting = 0
    peak = 0
    for thermal_zone in self.thermal_zones_from_internal_zones:
      lighting = thermal_zone.lighting
      for schedule in lighting.schedules:
        peak = max(schedule.values) * lighting.density * thermal_zone.total_floor_area
        if peak > peak_lighting:
          peak_lighting = peak
    results[cte.MONTH] = [peak for _ in range(0, 12)]
    results[cte.YEAR] = [peak]
    return results

  @property
  def appliances_peak_load(self) -> Union[None, dict]:
    """
    Get appliances peak load in W
    :return: dict{[float]}
    """
    results = {}
    peak_appliances = 0
    peak = 0
    for thermal_zone in self.thermal_zones_from_internal_zones:
      appliances = thermal_zone.appliances
      for schedule in appliances.schedules:
        peak = max(schedule.values) * appliances.density * thermal_zone.total_floor_area
        if peak > peak_appliances:
          peak_appliances = peak
    results[cte.MONTH] = [peak for _ in range(0, 12)]
    results[cte.YEAR] = [peak]
    return results

  @property
  def heating_peak_load(self) -> Union[None, dict]:
    """
    Get heating peak load in W
    :return: dict{[float]}
    """
    results = {}
    if cte.HOUR in self.heating_demand:
      monthly_values = PeakLoads().peak_loads_from_hourly(self.heating_demand[cte.HOUR])
    else:
      monthly_values = PeakLoads(self).heating_peak_loads_from_methodology
    if monthly_values is None:
      return None
    results[cte.MONTH] = [x * cte.WATTS_HOUR_TO_JULES for x in monthly_values]
    results[cte.YEAR] = [max(monthly_values)]
    return results

  @property
  def cooling_peak_load(self) -> Union[None, dict]:
    """
    Get cooling peak load in W
    :return: dict{[float]}
    """
    results = {}
    if cte.HOUR in self.cooling_demand:
      monthly_values = PeakLoads().peak_loads_from_hourly(self.cooling_demand[cte.HOUR])
    else:
      monthly_values = PeakLoads(self).cooling_peak_loads_from_methodology
    if monthly_values is None:
      return None
    results[cte.MONTH] = [x * cte.WATTS_HOUR_TO_JULES for x in monthly_values]
    results[cte.YEAR] = [max(monthly_values)]
    return results

  @property
  def eave_height(self):
    """
    Get building eave height in meters
    :return: float
    """
    if self._eave_height is None:
      self._eave_height = 0
      for wall in self.walls:
        self._eave_height = max(self._eave_height, wall.upper_corner[2]) - self.simplified_polyhedron.min_z
    return self._eave_height

  @property
  def roof_type(self):
    """
    Get roof type for the building flat or pitch
    :return: str
    """
    if self._roof_type is None:
      self._roof_type = 'flat'
      for roof in self.roofs:
        grads = np.rad2deg(roof.inclination)
        if 355 > grads > 5:
          self._roof_type = 'pitch'
          break
    return self._roof_type

  @roof_type.setter
  def roof_type(self, value):
    """
    Set roof type for the building flat or pitch
    :return: str
    """
    self._roof_type = value

  @property
  def floor_area(self):
    """
    Get building floor area in square meters
    :return: float
    """
    if self._floor_area is None:
      self._floor_area = 0
      for surface in self.surfaces:
        if surface.type == 'Ground':
          self._floor_area += surface.perimeter_polygon.area
    return self._floor_area

  @property
  def households(self) -> List[Household]:
    """
    Get the list of households inside the building
    :return: List[Household]
    """
    return self._households

  @property
  def is_conditioned(self):
    """
    Get building heated flag
    :return: Boolean
    """
    if self.internal_zones is None:
      return False
    for internal_zone in self.internal_zones:
      if internal_zone.usages is not None:
        for usage in internal_zone.usages:
          if usage.thermal_control is not None:
            return True
    return False

  @property
  def aliases(self):
    """
    Get the alias name for the building
    :return: str
    """
    return self._aliases

  def add_alias(self, value):
    """
    Add a new alias for the building
    """
    self._aliases.append(value)
    if self.city is not None:
      self.city.add_building_alias(self, value)

  @property
  def city(self) -> City:
    """
    Get the city containing the building
    :return: City
    """
    return self._city

  @city.setter
  def city(self, value):
    """
    Set the city containing the building
    """
    self._city = value

  @property
  def usages_percentage(self):
    """
    Get the usages and percentages for the building
    """
    _usage = ''
    for internal_zone in self.internal_zones:
      if internal_zone.usages is None:
        continue
      for usage in internal_zone.usages:
        _usage = f'{_usage}{usage.name}_{usage.percentage} '
    return _usage.rstrip()

  @property
  def energy_systems(self) -> Union[None, List[EnergySystem]]:
    """
    Get list of energy systems installed to cover the building demands
    :return: [EnergySystem]
    """
    return self._energy_systems

  @energy_systems.setter
  def energy_systems(self, value):
    """
    Set list of energy systems installed to cover the building demands
    :param value: [EnergySystem]
    """
    self._energy_systems = value

  @property
  def energy_systems_archetype_name(self):
    """
    Get energy systems archetype name
    :return: str
    """
    return self._systems_archetype_name

  @energy_systems_archetype_name.setter
  def energy_systems_archetype_name(self, value):
    """
    Set energy systems archetype name
    :param value: str
    """
    self._systems_archetype_name = value

  @property
  def heating_consumption(self):
    """
    Get energy consumption for heating according to the heating system installed in J
    return: dict
    """
    if len(self._heating_consumption) == 0:
      for heating_demand_key in self.heating_demand:
        demand = self.heating_demand[heating_demand_key]
        consumption_type = cte.HEATING
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._heating_consumption[heating_demand_key] = final_energy_consumed
    return self._heating_consumption

  @property
  def cooling_consumption(self):
    """
    Get energy consumption for cooling according to the cooling system installed in J
    return: dict
    """
    if len(self._cooling_consumption) == 0:
      for cooling_demand_key in self.cooling_demand:
        demand = self.cooling_demand[cooling_demand_key]
        consumption_type = cte.COOLING
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._cooling_consumption[cooling_demand_key] = final_energy_consumed
    return self._cooling_consumption

  @property
  def domestic_hot_water_consumption(self):
    """
    Get energy consumption for domestic according to the domestic hot water system installed in J
    return: dict
    """
    if len(self._domestic_hot_water_consumption) == 0:
      for domestic_hot_water_demand_key in self.domestic_hot_water_heat_demand:
        demand = self.domestic_hot_water_heat_demand[domestic_hot_water_demand_key]
        consumption_type = cte.DOMESTIC_HOT_WATER
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._domestic_hot_water_consumption[domestic_hot_water_demand_key] = final_energy_consumed
    return self._domestic_hot_water_consumption

  def _calculate_working_hours(self):
    _working_hours = {}
    for internal_zone in self.internal_zones:
      for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
        _working_hours_per_thermal_zone = {}
        for schedule in thermal_zone.thermal_control.hvac_availability_schedules:
          _working_hours_per_schedule = [0] * len(schedule.values)
          for i, value in enumerate(schedule.values):
            if value > 0:
              _working_hours_per_schedule[i] = 1
          for day_type in schedule.day_types:
            _working_hours_per_thermal_zone[day_type] = _working_hours_per_schedule
        if len(_working_hours) == 0:
          _working_hours = _working_hours_per_thermal_zone
        else:
          for key, item in _working_hours.items():
            saved_values = _working_hours_per_thermal_zone[key]
            for i, value in enumerate(item):
              _working_hours[key][i] = max(_working_hours[key][i], saved_values[i])

    _total_hours = 0
    for key in _working_hours:
      hours = sum(_working_hours[key])
      _total_hours += hours * cte.WEEK_DAYS_A_YEAR[key]
    return _total_hours

  @property
  def distribution_systems_electrical_consumption(self):
    """
    Get total electricity consumption for distribution and emission systems in J
    return: dict
    """
    if len(self._distribution_systems_electrical_consumption) != 0:
      return self._distribution_systems_electrical_consumption
    _peak_load = self.heating_peak_load[cte.YEAR][0]
    _peak_load_type = cte.HEATING
    if _peak_load < self.cooling_peak_load[cte.YEAR][0]:
      _peak_load = self.cooling_peak_load[cte.YEAR][0]
      _peak_load_type = cte.COOLING

    _working_hours = self._calculate_working_hours()
    _consumption_fix_flow = 0
    if self.energy_systems is None:
      return self._distribution_systems_electrical_consumption
    for energy_system in self.energy_systems:
      emission_system = energy_system.emission_system.generic_emission_system
      parasitic_energy_consumption = 0
      if emission_system is not None:
        parasitic_energy_consumption = emission_system.parasitic_energy_consumption
      distribution_system = energy_system.distribution_system.generic_distribution_system
      consumption_variable_flow = distribution_system.distribution_consumption_variable_flow
      for demand_type in energy_system.demand_types:
        if demand_type.lower() == cte.HEATING.lower():
          if _peak_load_type == cte.HEATING.lower():
            _consumption_fix_flow = distribution_system.distribution_consumption_fix_flow
          for heating_demand_key in self.heating_demand:
            _consumption = [0]*len(self.heating_demand[heating_demand_key])
            _demand = self.heating_demand[heating_demand_key]
            for i, _ in enumerate(_consumption):
              _consumption[i] += (parasitic_energy_consumption + consumption_variable_flow) * _demand[i]
            self._distribution_systems_electrical_consumption[heating_demand_key] = _consumption
        if demand_type.lower() == cte.COOLING.lower():
          if _peak_load_type == cte.COOLING.lower():
            _consumption_fix_flow = distribution_system.distribution_consumption_fix_flow
          for demand_key in self.cooling_demand:
            _consumption = self._distribution_systems_electrical_consumption[demand_key]
            _demand = self.cooling_demand[demand_key]
            for i, _ in enumerate(_consumption):
              _consumption[i] += (parasitic_energy_consumption + consumption_variable_flow) * _demand[i]
            self._distribution_systems_electrical_consumption[demand_key] = _consumption

      for key, item in self._distribution_systems_electrical_consumption.items():
        for i in range(0, len(item)):
          self._distribution_systems_electrical_consumption[key][i] += _peak_load * _consumption_fix_flow \
                                                                       * _working_hours
    return self._distribution_systems_electrical_consumption

  def _calculate_consumption(self, consumption_type, demand):
    # todo: modify when COP depends on the hour
    coefficient_of_performance = 0
    if self.energy_systems is None:
      return None
    for energy_system in self.energy_systems:
      for demand_type in energy_system.demand_types:
        if demand_type.lower() == consumption_type.lower():
          if consumption_type in (cte.HEATING, cte.DOMESTIC_HOT_WATER):
            coefficient_of_performance = energy_system.generation_system.generic_generation_system.heat_efficiency
          elif consumption_type == cte.COOLING:
            coefficient_of_performance = energy_system.generation_system.generic_generation_system.cooling_efficiency
          elif consumption_type == cte.ELECTRICITY:
            coefficient_of_performance = \
              energy_system.generation_system.generic_generation_system.electricity_efficiency
    if coefficient_of_performance == 0:
      values = [0]*len(demand)
      final_energy_consumed = values
    else:
      final_energy_consumed = []
      for demand_value in demand:
        final_energy_consumed.append(demand_value / coefficient_of_performance)
    return final_energy_consumed

  @property
  def onsite_electrical_production(self):
    """
    Get total electricity produced onsite in J
    return: dict
    """
    orientation_losses_factor = {cte.MONTH: {'north': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                             'east': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                             'south': [2.137931, 1.645503, 1.320946, 1.107817, 0.993213, 0.945175,
                                                       0.967949, 1.065534, 1.24183, 1.486486, 1.918033, 2.210526],
                                             'west': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]},
                                 cte.YEAR: {'north': [0],
                                            'east': [0],
                                            'south': [1.212544],
                                            'west': [0]}
                                 }

    # Add other systems whenever new ones appear
    if self.energy_systems is None:
      return self._onsite_electrical_production
    for energy_system in self.energy_systems:
      if energy_system.generation_system.generic_generation_system.type == cte.PHOTOVOLTAIC:
        _efficiency = energy_system.generation_system.generic_generation_system.electricity_efficiency
        self._onsite_electrical_production = {}
        for _key in self.roofs[0].global_irradiance.keys():
          _results = [0 for _ in range(0, len(self.roofs[0].global_irradiance[_key]))]
          for surface in self.roofs:
            if _key in orientation_losses_factor:
              _results = [x + y * _efficiency * surface.perimeter_area
                          * surface.solar_collectors_area_reduction_factor * z
                          for x, y, z in zip(_results, surface.global_irradiance[_key],
                                             orientation_losses_factor[_key]['south'])]
          self._onsite_electrical_production[_key] = _results
    return self._onsite_electrical_production



"""
InselMonthlyEnergyBalance exports models to insel format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging
from pathlib import Path

import numpy as np

import hub.helpers.constants as cte
from hub.imports.weather.helpers.weather import Weather

_CONSTRUCTION_CODE = {
  cte.WALL: '1',
  cte.GROUND: '2',
  cte.ROOF: '3',
  cte.INTERIOR_WALL: '5',
  cte.GROUND_WALL: '6',
  cte.ATTIC_FLOOR: '7',
  cte.INTERIOR_SLAB: '8'
}

_NUMBER_DAYS_PER_MONTH = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]


class InselMonthlyEnergyBalance:
  """
  Insel monthly energy balance class
  """
  def __init__(self, city, path, custom_insel_block):
    self._city = city
    self._path = path
    self._custom_insel_block = custom_insel_block
    self._results = None
    self._contents = []
    self._insel_files_paths = []
    self._sanity_check()
    for building in city.buildings:
      self._insel_files_paths.append(building.name + '.insel')
      file_name_out = building.name + '.out'
      output_path = Path(self._path / file_name_out).resolve()
      if building.thermal_zones_from_internal_zones is None:
        logging.warning('Building %s has missing values. Monthly Energy Balance cannot be processed', building.name)

      self._contents.append(
        self._generate_meb_template(building, output_path, self._custom_insel_block)
      )
    self._export()

  @staticmethod
  def _add_block(file, block_number, block_type, inputs=None, parameters=None):
    file += "S " + str(block_number) + " " + block_type + "\n"
    if inputs is not None:
      for block_input in inputs:
        file += str(block_input) + "\n"
    if parameters is not None:
      if len(parameters) > 0:
        file += "P " + str(block_number) + "\n"
      for block_parameter in parameters:
        file += str(block_parameter) + "\n"
    return file

  def _export(self):
    for i_file, content in enumerate(self._contents):
      file_name = self._insel_files_paths[i_file]
      with open(Path(self._path / file_name).resolve(), 'w', encoding='utf8') as insel_file:
        insel_file.write(content)

  def _sanity_check(self):
    levels_of_detail = self._city.level_of_detail
    if levels_of_detail.geometry is None:
      raise AttributeError('Level of detail of geometry not assigned')
    if levels_of_detail.geometry < 1:
      raise AttributeError(f'Level of detail of geometry = {levels_of_detail.geometry}. Required minimum level 1')
    if levels_of_detail.construction is None:
      raise AttributeError('Level of detail of construction not assigned')
    if levels_of_detail.construction < 1:
      raise AttributeError(f'Level of detail of construction = {levels_of_detail.construction}. Required minimum level 1')
    if levels_of_detail.usage is None:
      raise AttributeError('Level of detail of usage not assigned')
    if levels_of_detail.usage < 1:
      raise AttributeError(f'Level of detail of usage = {levels_of_detail.usage}. Required minimum level 1')
    if levels_of_detail.weather is None:
      raise AttributeError('Level of detail of weather not assigned')
    if levels_of_detail.weather < 1:
      raise AttributeError(f'Level of detail of weather = {levels_of_detail.weather}. Required minimum level 1')
    if levels_of_detail.surface_radiation is None:
      raise AttributeError('Level of detail of surface radiation not assigned')
    if levels_of_detail.surface_radiation < 1:
      raise AttributeError(f'Level of detail of surface radiation = {levels_of_detail.surface_radiation}. '
                           f'Required minimum level 1')

  @staticmethod
  def _generate_meb_template(building, insel_outputs_path, custom_insel_block):
    file = ""
    i_block = 1
    parameters = ["1", "12", "1"]
    file = InselMonthlyEnergyBalance._add_block(file, i_block, 'DO', parameters=parameters)

    i_block = 4
    inputs = ["1.1", "20.1", "21.1"]
    surfaces = building.surfaces
    for i in range(1, len(surfaces) + 1):
      inputs.append(f"{str(100 + i)}.1 % Radiation surface {str(i)}")

    number_of_storeys = int(building.eave_height / building.average_storey_height)
    attic_heated = building.attic_heated
    basement_heated = building.basement_heated
    if building.attic_heated is None:
      attic_heated = 0
    if building.basement_heated is None:
      basement_heated = 0

    # BUILDING PARAMETERS
    parameters = [f'{building.volume} % BP(1) Heated Volume (m3)',
                  f'{building.average_storey_height} % BP(2) Average storey height (m)',
                  f'{number_of_storeys} % BP(3) Number of storeys above ground',
                  f'{attic_heated} % BP(4) Attic heating type (0=no room, 1=unheated, 2=heated)',
                  f'{basement_heated} % BP(5) Cellar heating type (0=no room, 1=unheated, 2=heated, '
                  f'99=invalid)']

    # todo: this method and the insel model have to be reviewed for more than one internal zone
    internal_zone = building.internal_zones[0]
    thermal_zone = internal_zone.thermal_zones_from_internal_zones[0]
    parameters.append(f'{thermal_zone.indirectly_heated_area_ratio} % BP(6) Indirectly heated area ratio')
    parameters.append(f'{thermal_zone.effective_thermal_capacity / 3600 / building.average_storey_height}'
                      f' % BP(7) Effective heat capacity (Wh/m2K)')
    parameters.append(f'{thermal_zone.additional_thermal_bridge_u_value} '
                      f'% BP(8) Additional U-value for heat bridge (W/m2K)')
    parameters.append('1 % BP(9) Usage type (0=standard, 1=IWU)')

    # ZONES AND SURFACES
    parameters.append(f'{len(internal_zone.usages)} %  BP(10) Number of zones')

    for i, usage in enumerate(internal_zone.usages):
      percentage_usage = usage.percentage
      parameters.append(f'{internal_zone.thermal_zones_from_internal_zones[0].total_floor_area * percentage_usage} '
                        f'% BP(11) #1 Area of zone {i + 1} (m2)')
      total_internal_gain = 0
      for i_gain in usage.internal_gains:
        internal_gain = i_gain.average_internal_gain * (i_gain.convective_fraction + i_gain.radiative_fraction)
        for schedule in i_gain.schedules:
          total_values = sum(schedule.values)
          total_hours = 0
          for day_type in schedule.day_types:
            total_hours += cte.WEEK_DAYS_A_YEAR[day_type] / 365 / 24
          total_values *= total_hours
          total_internal_gain += internal_gain * total_values

      parameters.append(f'{total_internal_gain} % BP(12) #2 Internal gains of zone {i + 1}')
      parameters.append(f'{usage.thermal_control.mean_heating_set_point} % BP(13) #3 Heating setpoint temperature '
                        f'zone {i + 1} (degree Celsius)')
      parameters.append(f'{usage.thermal_control.heating_set_back} % BP(14) #4 Heating setback temperature '
                        f'zone {i + 1} (degree Celsius)')
      parameters.append(f'{usage.thermal_control.mean_cooling_set_point} % BP(15) #5 Cooling setpoint temperature '
                        f'zone {i + 1} (degree Celsius)')
      parameters.append(f'{usage.hours_day} %  BP(16) #6 Usage hours per day zone {i + 1}')
      parameters.append(f'{usage.days_year} %  BP(17) #7 Usage days per year zone {i + 1}')

      ventilation = 0
      infiltration = 0
      for schedule in usage.thermal_control.hvac_availability_schedules:
        ventilation_day = 0
        infiltration_day = 0
        for value in schedule.values:
          if value == 0:
            infiltration_day += internal_zone.thermal_zones_from_internal_zones[0].infiltration_rate_system_off / 24 * cte.HOUR_TO_SECONDS
            ventilation_day += 0
          else:
            ventilation_value = usage.mechanical_air_change * value * cte.HOUR_TO_SECONDS
            infiltration_value = internal_zone.thermal_zones_from_internal_zones[0].infiltration_rate_system_off * value * cte.HOUR_TO_SECONDS
            if ventilation_value >= infiltration_value:
              ventilation_day += ventilation_value / 24
              infiltration_day += 0
            else:
              ventilation_day += 0
              infiltration_day += infiltration_value / 24
        for day_type in schedule.day_types:
          infiltration += infiltration_day * cte.WEEK_DAYS_A_YEAR[day_type] / 365
          ventilation += ventilation_day * cte.WEEK_DAYS_A_YEAR[day_type] / 365

      ventilation_infiltration = ventilation + infiltration
      parameters.append(f'{ventilation_infiltration} % BP(18) #8 Minimum air change rate zone {i + 1} (ACH)')

    parameters.append(f'{len(thermal_zone.thermal_boundaries)}  % Number of surfaces = BP(11+8z) \n'
                      f'% 1. Surface type (1=wall, 2=ground 3=roof, 4=flat roof)\n'
                      f'% 2. Areas above ground (m2)\n'
                      f'% 3. Areas below ground (m2)\n'
                      f'% 4. U-value (W/m2K)\n'
                      f'% 5. Window area (m2)\n'
                      f'% 6. Window frame fraction\n'
                      f'% 7. Window U-value (W/m2K)\n'
                      f'% 8. Window g-value\n'
                      f'% 9. Short-wave reflectance\n'
                      f'% #1     #2       #3      #4      #5     #6     #7     #8     #9\n')

    for thermal_boundary in thermal_zone.thermal_boundaries:
      type_code = _CONSTRUCTION_CODE[thermal_boundary.type]
      wall_area = thermal_boundary.opaque_area * (1 + thermal_boundary.window_ratio)
      if thermal_boundary.type == cte.WALL:
        if thermal_boundary.parent_surface.percentage_shared is not None:
          wall_area = wall_area * (1 - thermal_boundary.parent_surface.percentage_shared)
      window_area = wall_area * thermal_boundary.window_ratio

      parameters.append(type_code)
      if thermal_boundary.type != cte.GROUND:
        parameters.append(wall_area)
        parameters.append('0.0')
      else:
        parameters.append('0.0')
        parameters.append(wall_area)
      parameters.append(thermal_boundary.u_value)
      parameters.append(window_area)

      if window_area <= 0.001:
        parameters.append(0.0)
        parameters.append(0.0)
        parameters.append(0.0)
      else:
        thermal_opening = thermal_boundary.thermal_openings[0]
        parameters.append(thermal_opening.frame_ratio)
        parameters.append(thermal_opening.overall_u_value)
        parameters.append(thermal_opening.g_value)
      if thermal_boundary.type is not cte.GROUND:
        parameters.append(thermal_boundary.external_surface.short_wave_reflectance)
      else:
        parameters.append(0.0)
    file = InselMonthlyEnergyBalance._add_block(file, i_block, custom_insel_block, inputs=inputs, parameters=parameters)

    i_block = 20
    inputs = ['1']
    parameters = ['12 % Monthly ambient temperature (degree Celsius)']

    external_temperature = building.external_temperature[cte.MONTH]
    for i in range(0, len(external_temperature)):
      parameters.append(f'{i + 1} {external_temperature[i]}')

    file = InselMonthlyEnergyBalance._add_block(file, i_block, 'polyg', inputs=inputs, parameters=parameters)

    i_block = 21
    inputs = ['1']
    parameters = ['12 % Monthly sky temperature']

    sky_temperature = Weather.sky_temperature(external_temperature)
    for i, temperature in enumerate(sky_temperature):
      parameters.append(f'{i + 1} {temperature}')

    file = InselMonthlyEnergyBalance._add_block(file, i_block, 'polyg', inputs=inputs, parameters=parameters)
    for i, surface in enumerate(surfaces):
      i_block = 101 + i
      inputs = ['1 % Monthly surface radiation (W/m2)']
      parameters = [f'12 % Azimuth {np.rad2deg(surface.azimuth)}, '
                    f'inclination {np.rad2deg(surface.inclination)} (degrees)']

      if surface.type != 'Ground':
        if cte.MONTH not in surface.global_irradiance:
          raise ValueError(f'surface: {surface.name} from building {building.name} has no global irradiance!')

        global_irradiance = surface.global_irradiance[cte.MONTH]
        for j in range(0, len(global_irradiance)):
          parameters.append(f'{j + 1} '
                            f'{global_irradiance[j] * cte.WATTS_HOUR_TO_JULES / 24 / _NUMBER_DAYS_PER_MONTH[j]}')
      else:
        for j in range(0, 12):
          parameters.append(f'{j + 1} 0.0')

      file = InselMonthlyEnergyBalance._add_block(file, i_block, 'polyg', inputs=inputs, parameters=parameters)

    i_block = 300 + len(surfaces)
    inputs = ['4.1', '4.2']
    file = InselMonthlyEnergyBalance._add_block(file, i_block, 'cum', inputs=inputs)

    in_1 = f'{i_block}.1'
    in_2 = f'{i_block}.2'
    i_block = 303 + len(surfaces)
    inputs = [in_1, in_2]
    file = InselMonthlyEnergyBalance._add_block(file, i_block, 'atend', inputs=inputs)

    i_block = 310 + len(surfaces)
    inputs = ['4.1', '4.2']
    parameters = ['1 % Mode',
                  '0 % Suppress FNQ inputs',
                  f"'{str(insel_outputs_path)}' % File name",
                  "'*' % Fortran format"]
    file = InselMonthlyEnergyBalance._add_block(file, i_block, 'WRITE', inputs=inputs, parameters=parameters)
    return file



"""
Energy System catalog distribution system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class DistributionSystem:
  """
  Distribution system class
  """

  def __init__(self, system_id, name, system_type, supply_temperature, distribution_consumption_fix_flow,
               distribution_consumption_variable_flow, heat_losses):
    self._system_id = system_id
    self._name = name
    self._type = system_type
    self._supply_temperature = supply_temperature
    self._distribution_consumption_fix_flow = distribution_consumption_fix_flow
    self._distribution_consumption_variable_flow = distribution_consumption_variable_flow
    self._heat_losses = heat_losses

  @property
  def id(self):
    """
    Get system id
    :return: float
    """
    return self._system_id

  @property
  def name(self):
    """
    Get name
    :return: string
    """
    return self._name

  @property
  def type(self):
    """
    Get type from [air, water, refrigerant]
    :return: string
    """
    return self._type

  @property
  def supply_temperature(self):
    """
    Get supply_temperature in degree Celsius
    :return: float
    """
    return self._supply_temperature

  @property
  def distribution_consumption_fix_flow(self):
    """
    Get distribution_consumption if the pump or fan work at fix mass or volume flow in ratio over peak power (W/W)
    :return: float
    """
    return self._distribution_consumption_fix_flow

  @property
  def distribution_consumption_variable_flow(self):
    """
    Get distribution_consumption if the pump or fan work at variable mass or volume flow in ratio
    over energy produced (J/J)
    :return: float
    """
    return self._distribution_consumption_variable_flow

  @property
  def heat_losses(self):
    """
    Get heat_losses in ratio over energy produced in J/J
    :return: float
    """
    return self._heat_losses

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {
      'Layer': {
        'id': self.id,
        'name': self.name,
        'type': self.type,
        'supply temperature [Celsius]': self.supply_temperature,
        'distribution consumption if fix flow over peak power [W/W]': self.distribution_consumption_fix_flow,
        'distribution consumption if variable flow over peak power [J/J]': self.distribution_consumption_variable_flow,
        'heat losses per energy produced [J/J]': self.heat_losses
      }
    }
    return content



"""
export a city into Stl format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.exports.formats.triangular import Triangular


class Stl(Triangular):
  """
  Export to STL
  """
  def __init__(self, city, path):
    super().__init__(city, path, 'stl', write_mode='wb')



"""
Energy distribution system definition
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.city_model_structure.energy_systems.generic_distribution_system import GenericDistributionSystem


class DistributionSystem:
  """
  DistributionSystem class
  """
  def __init__(self):
    self._generic_distribution_system = None

  @property
  def generic_distribution_system(self) -> GenericDistributionSystem:
    """
    Get generic_distribution_system
    :return: GenericDistributionSystem
    """
    return self._generic_distribution_system

  @generic_distribution_system.setter
  def generic_distribution_system(self, value):
    """
    Set associated generic_distribution_system
    :param value: GenericDistributionSystem
    """
    self._generic_distribution_system = value



"""
Cost chapter description
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List
from hub.catalog_factories.data_models.cost.item_description import ItemDescription


class Chapter:
  """
  Chapter class
  """
  def __init__(self, chapter_type, items):

    self._chapter_type = chapter_type
    self._items = items

  @property
  def chapter_type(self):
    """
    Get chapter type
    :return: str
    """
    return self._chapter_type

  @property
  def items(self) -> List[ItemDescription]:
    """
    Get list of items contained in the chapter
    :return: [str]
    """
    return self._items

  def item(self, name) -> ItemDescription:
    """
    Get specific item by name
    :return: ItemDescription
    """
    for item in self.items:
      if item.type == name:
        return item
    raise KeyError(f'Item name {name} not found')

  def to_dictionary(self):
    """Class content to dictionary"""
    _items = []
    for _item in self.items:
      _items.append(_item.to_dictionary())
    content = {'Chapter': {'chapter type': self.chapter_type,
                           'items': _items
                           }
               }

    return content



"""
Plant module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List
from hub.city_model_structure.greenery.soil import Soil


class Plant:
  """
  Plant class
  """
  def __init__(self, name, height, leaf_area_index, leaf_reflectivity, leaf_emissivity, minimal_stomatal_resistance,
               co2_sequestration, grows_on_soils):
    self._name = name
    self._percentage = None
    self._height = height
    self._leaf_area_index = leaf_area_index
    self._leaf_reflectivity = leaf_reflectivity
    self._leaf_emissivity = leaf_emissivity
    self._minimal_stomatal_resistance = minimal_stomatal_resistance
    self._co2_sequestration = co2_sequestration
    self._grows_on = grows_on_soils

  @property
  def name(self):
    """
    Get plant name
    :return: string
    """
    return self._name

  @property
  def percentage(self):
    """
    Get percentage of plant in vegetation
    :return: float
    """
    return self._percentage

  @percentage.setter
  def percentage(self, value):
    """
    Set percentage of plant in vegetation
    :param value: float
    """
    self._percentage = value

  @property
  def height(self):
    """
    Get plant height in m
    :return: float
    """
    return self._height

  @property
  def leaf_area_index(self):
    """
    Get plant leaf area index
    :return: float
    """
    return self._leaf_area_index

  @property
  def leaf_reflectivity(self):
    """
    Get plant leaf area index
    :return: float
    """
    return self._leaf_reflectivity

  @property
  def leaf_emissivity(self):
    """
    Get plant leaf emissivity
    :return: float
    """
    return self._leaf_emissivity

  @property
  def minimal_stomatal_resistance(self):
    """
    Get plant minimal stomatal resistance in s/m
    :return: float
    """
    return self._minimal_stomatal_resistance

  @property
  def co2_sequestration(self):
    """
    Get plant co2 sequestration capacity in kg CO2 equivalent
    :return: float
    """
    return self._co2_sequestration

  @property
  def grows_on(self) -> List[Soil]:
    """
    Get plant compatible soils
    :return: [Soil]
    """
    return self._grows_on



"""
export a city from trimesh into Triangular format (obj or stl)
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
from pathlib import Path
from trimesh import Trimesh


class Triangular:
  """
  Superclass to export to triangular format (STL or OBJ)
  """
  def __init__(self, city, path, triangular_format, write_mode='w'):
    self._city = city
    self._path = path
    self._triangular_format = triangular_format
    self._write_mode = write_mode
    self._export()

  def _export(self):
    if self._city.name is None:
      self._city.name = 'unknown_city'
    file_name = self._city.name + '.' + self._triangular_format
    file_path = (Path(self._path).resolve() / file_name).resolve()
    trimesh = Trimesh()
    for building in self._city.buildings:
      trimesh = trimesh.union(building.simplified_polyhedron.trimesh)

    with open(file_path, self._write_mode) as file:
      file.write(trimesh.export(file_type=self._triangular_format))



"""
Calculation of loads for peak heating and cooling
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class LoadsCalculation:
  """
  LoadsCalculation class
  """

  def __init__(self, building):
    self._building = building

  @staticmethod
  def _get_load_transmitted(thermal_zone, internal_temperature, ambient_temperature, ground_temperature):
    load_transmitted_opaque = 0
    load_transmitted_transparent = 0
    for thermal_boundary in thermal_zone.thermal_boundaries:
      if thermal_boundary.type == cte.GROUND:
        external_temperature = ground_temperature
      elif thermal_boundary.type == cte.INTERIOR_WALL:
        external_temperature = internal_temperature
      else:
        external_temperature = ambient_temperature

      load_transmitted_opaque += (
          thermal_boundary.u_value * thermal_boundary.opaque_area * (internal_temperature - external_temperature)
      )
      for thermal_opening in thermal_boundary.thermal_openings:
        load_transmitted_transparent += thermal_opening.overall_u_value \
                                        * (internal_temperature - external_temperature)
    load_transmitted_opaque += (
        thermal_zone.additional_thermal_bridge_u_value * thermal_zone.footprint_area *
        (internal_temperature - ambient_temperature)
    )
    load_transmitted = load_transmitted_opaque + load_transmitted_transparent
    return load_transmitted

  @staticmethod
  def _get_load_ventilation(thermal_zone, internal_temperature, ambient_temperature):
    load_renovation_sensible = 0
    for usage in thermal_zone.usages:
      load_renovation_sensible += cte.AIR_DENSITY * cte.AIR_HEAT_CAPACITY * usage.mechanical_air_change \
                                  * thermal_zone.volume * (internal_temperature - ambient_temperature)

    load_infiltration_sensible = (
        cte.AIR_DENSITY * cte.AIR_HEAT_CAPACITY * thermal_zone.infiltration_rate_system_off * thermal_zone.volume
        * (internal_temperature - ambient_temperature)
    )

    load_ventilation = load_renovation_sensible + load_infiltration_sensible

    return load_ventilation

  def get_heating_transmitted_load(self, ambient_temperature, ground_temperature):
    """
    Calculates the heating transmitted load
    :return: int
    """
    heating_load_transmitted = 0
    for thermal_zone in self._building.thermal_zones_from_internal_zones:
      internal_temperature = thermal_zone.thermal_control.mean_heating_set_point
      heating_load_transmitted += self._get_load_transmitted(thermal_zone, internal_temperature, ambient_temperature,
                                                             ground_temperature)
    return heating_load_transmitted

  def get_cooling_transmitted_load(self, ambient_temperature, ground_temperature):
    """
    Calculates the cooling transmitted load
    :return: int
    """
    cooling_load_transmitted = 0
    for thermal_zone in self._building.thermal_zones_from_internal_zones:
      internal_temperature = thermal_zone.thermal_control.mean_cooling_set_point
      cooling_load_transmitted += self._get_load_transmitted(thermal_zone, internal_temperature, ambient_temperature,
                                                             ground_temperature)
    return cooling_load_transmitted

  def get_heating_ventilation_load_sensible(self, ambient_temperature):
    """
    Calculates the heating ventilation load sensible
    :return: int
    """
    heating_ventilation_load = 0
    for thermal_zone in self._building.thermal_zones_from_internal_zones:
      internal_temperature = thermal_zone.thermal_control.mean_heating_set_point
      heating_ventilation_load += self._get_load_ventilation(thermal_zone, internal_temperature, ambient_temperature)
    return heating_ventilation_load

  def get_cooling_ventilation_load_sensible(self, ambient_temperature):
    """
    Calculates the cooling ventilation load sensible
    :return: int
    """
    cooling_ventilation_load = 0
    for thermal_zone in self._building.thermal_zones_from_internal_zones:
      internal_temperature = thermal_zone.thermal_control.mean_cooling_set_point
      cooling_ventilation_load += self._get_load_ventilation(thermal_zone, internal_temperature, ambient_temperature)
    return cooling_ventilation_load

  def get_internal_load_sensible(self):
    """
    Calculates the internal load sensible
    :return: int
    """
    cooling_load_occupancy_sensible = 0
    cooling_load_lighting = 0
    cooling_load_equipment_sensible = 0
    for thermal_zone in self._building.thermal_zones_from_internal_zones:
      cooling_load_occupancy_sensible += (thermal_zone.occupancy.sensible_convective_internal_gain
                                          + thermal_zone.occupancy.sensible_radiative_internal_gain) \
                                         * thermal_zone.footprint_area
      cooling_load_lighting += (
          thermal_zone.lighting.density * thermal_zone.lighting.convective_fraction + thermal_zone.lighting.density *
          thermal_zone.lighting.radiative_fraction
      ) * thermal_zone.footprint_area
      cooling_load_equipment_sensible += (
          thermal_zone.appliances.density * thermal_zone.appliances.convective_fraction +
          thermal_zone.appliances.density * thermal_zone.appliances.radiative_fraction
      ) * thermal_zone.footprint_area
    internal_load = cooling_load_occupancy_sensible + cooling_load_lighting + cooling_load_equipment_sensible
    return internal_load

  def get_radiation_load(self, hour):
    """
    Calculates the radiation load
    :return: int
    """
    cooling_load_radiation = 0
    for thermal_zone in self._building.thermal_zones_from_internal_zones:
      for thermal_boundary in thermal_zone.thermal_boundaries:
        for thermal_opening in thermal_boundary.thermal_openings:
          radiation = thermal_boundary.parent_surface.global_irradiance[cte.HOUR][hour] * cte.WATTS_HOUR_TO_JULES
          cooling_load_radiation += (
              thermal_opening.area * (1 - thermal_opening.frame_ratio) * thermal_opening.g_value * radiation
          )
    return cooling_load_radiation



"""
Constant module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import logging


def validate_import_export_type(cls_name: type, handler: str):
  """
  Retrieves all the function names in a class which are property types (decoration)
  and normal functions
  :param cls_name: the class name
  :param handler: import export handler
  :return: None
  """
  functions = [
    function[1:] for function in dir(cls_name)
    if (
           isinstance(getattr(cls_name, function), property) or callable(getattr(cls_name, function))
       ) and function in cls_name.__dict__ and function[0] == '_' and function != '__init__']

  if handler.lower() not in functions:
    error_message = f'Wrong import type [{handler}]. Valid functions include {functions}'
    logging.error(error_message)
    raise ValueError(error_message)



"""
Dictionaries module for hub function to eilat construction function
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class HubFunctionToEilatConstructionFunction:
  """
  Hub function to Eilat construction function class
  """
  def __init__(self):
    self._dictionary = {
      cte.RESIDENTIAL: 'Residential_building',
      cte.HOTEL: 'Residential_building',
      cte.DORMITORY: 'Residential_building'
    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
Usage catalog usage
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
from hub.catalog_factories.data_models.usages.usage import Usage


class Content:
  """
  Content class
  """
  def __init__(self, usages):
    self._usages = usages

  @property
  def usages(self) -> [Usage]:
    """
    Get catalog usages
    """
    return self._usages

  def to_dictionary(self):
    """Class content to dictionary"""
    _usages = []
    for _usage in self.usages:
      _usages.append(_usage.to_dictionary())
    content = {'Usages': _usages}

    return content

  def __str__(self):
    """Print content"""
    _usages = []
    for _usage in self.usages:
      _usages.append(_usage.to_dictionary())
    content = {'Usages': _usages}

    return str(content)



"""
Catalog base class
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class Catalog:
  """
  Catalogs base class
  catalog_factories will inherit from this class.
  """

  def names(self, category=None):
    """
    Base property to return the catalog entries names.
    :return: Catalog names filter by category if provided
    """
    raise NotImplementedError

  def entries(self, category=None):
    """
    Base property to return the catalog entries
    :return: Catalog content filter by category if provided
    """
    raise NotImplementedError

  def get_entry(self, name):
    """
    Base property to return the catalog entry matching the given name
    :return: Catalog entry with the matching name
    """
    raise NotImplementedError



"""
ThermalZone module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import uuid
import copy
from typing import List, Union, TypeVar
import numpy

from hub.city_model_structure.building_demand.occupancy import Occupancy
from hub.city_model_structure.building_demand.appliances import Appliances
from hub.city_model_structure.building_demand.lighting import Lighting
from hub.city_model_structure.building_demand.internal_gain import InternalGain
from hub.city_model_structure.building_demand.thermal_control import ThermalControl
from hub.city_model_structure.building_demand.domestic_hot_water import DomesticHotWater
from hub.city_model_structure.attributes.schedule import Schedule
import hub.helpers.constants as cte

ThermalBoundary = TypeVar('ThermalBoundary')
InternalZone = TypeVar('InternalZone')


class ThermalZone:
  """
  ThermalZone class
  """

  def __init__(self, thermal_boundaries,
               parent_internal_zone,
               volume,
               footprint_area,
               number_of_storeys,
               usage_name=None):
    self._id = None
    self._parent_internal_zone = parent_internal_zone
    self._footprint_area = footprint_area
    self._thermal_boundaries = thermal_boundaries
    self._additional_thermal_bridge_u_value = None
    self._effective_thermal_capacity = None
    self._indirectly_heated_area_ratio = None
    self._infiltration_rate_system_on = None
    self._infiltration_rate_system_off = None
    self._volume = volume
    self._ordinate_number = None
    self._view_factors_matrix = None
    self._total_floor_area = None
    self._number_of_storeys = number_of_storeys
    self._usage_name = usage_name
    self._usage_from_parent = False
    if usage_name is None:
      self._usage_from_parent = True
    self._hours_day = None
    self._days_year = None
    self._mechanical_air_change = None
    self._occupancy = None
    self._lighting = None
    self._appliances = None
    self._internal_gains = None
    self._thermal_control = None
    self._domestic_hot_water = None
    self._usages = None

  @property
  def parent_internal_zone(self) -> InternalZone:
    """
    Get the internal zone to which this thermal zone belongs
    :return: InternalZone
    """
    return self._parent_internal_zone

  @property
  def usages(self):
    """
    Get the thermal zone usages including percentage with the format [percentage]-usage_[percentage]-usage...
    Eg: 70-office_30-residential
    :return: str
    """
    if self._usage_from_parent:
      self._usages = copy.deepcopy(self._parent_internal_zone.usages)
    else:
      values = self._usage_name.split('_')
      usages = []
      for value in values:
        usages.append(value.split('-'))
      self._usages = []
      for parent_usage in self._parent_internal_zone.usages:
        for value in usages:
          if parent_usage.name == value[1]:
            new_usage = copy.deepcopy(parent_usage)
            new_usage.percentage = float(value[0]) / 100
            self._usages.append(new_usage)
    return self._usages

  @property
  def id(self):
    """
    Get thermal zone id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def footprint_area(self) -> float:
    """
    Get thermal zone footprint area in m2
    :return: float
    """
    return self._footprint_area

  @property
  def thermal_boundaries(self) -> [ThermalBoundary]:
    """
    Get thermal boundaries bounding with the thermal zone
    :return: [ThermalBoundary]
    """
    return self._thermal_boundaries

  @property
  def additional_thermal_bridge_u_value(self) -> Union[None, float]:
    """
    Get thermal zone additional thermal bridge u value per footprint area W/m2K
    :return: None or float
    """
    self._additional_thermal_bridge_u_value = self.parent_internal_zone.thermal_archetype.extra_loses_due_to_thermal_bridges
    return self._additional_thermal_bridge_u_value

  @property
  def effective_thermal_capacity(self) -> Union[None, float]:
    """
    Get thermal zone effective thermal capacity in J/m3K
    :return: None or float
    """
    self._effective_thermal_capacity = self._parent_internal_zone.thermal_archetype.thermal_capacity
    return self._effective_thermal_capacity

  @property
  def indirectly_heated_area_ratio(self) -> Union[None, float]:
    """
    Get thermal zone indirectly heated area ratio
    :return: None or float
    """
    self._indirectly_heated_area_ratio = self._parent_internal_zone.thermal_archetype.indirect_heated_ratio
    return self._indirectly_heated_area_ratio

  @property
  def infiltration_rate_system_on(self):
    """
    Get thermal zone infiltration rate system on in air changes per second (1/s)
    :return: None or float
    """
    self._infiltration_rate_system_on = self._parent_internal_zone.thermal_archetype.infiltration_rate_for_ventilation_system_on
    return self._infiltration_rate_system_on

  @property
  def infiltration_rate_system_off(self):
    """
    Get thermal zone infiltration rate system off in air changes per second (1/s)
    :return: None or float
    """
    self._infiltration_rate_system_off = self._parent_internal_zone.thermal_archetype.infiltration_rate_for_ventilation_system_off
    return self._infiltration_rate_system_off

  @property
  def volume(self):
    """
    Get thermal zone volume
    :return: float
    """
    return self._volume

  @property
  def ordinate_number(self) -> Union[None, int]:
    """
    Get the order in which the thermal_zones need to be enumerated
    :return: None or int
    """
    return self._ordinate_number

  @ordinate_number.setter
  def ordinate_number(self, value):
    """
    Set a specific order of the zones to be called
    :param value: int
    """
    if value is not None:
      self._ordinate_number = int(value)

  @property
  def view_factors_matrix(self):
    """
    Get thermal zone view factors matrix
    :return: [[float]]
    """
    # todo: review method if windows not in window_ratio but in geometry
    if self._view_factors_matrix is None:
      total_area = 0
      for thermal_boundary in self.thermal_boundaries:
        total_area += thermal_boundary.opaque_area
        for thermal_opening in thermal_boundary.thermal_openings:
          total_area += thermal_opening.area

      view_factors_matrix = []
      for thermal_boundary_1 in self.thermal_boundaries:
        values = []
        for thermal_boundary_2 in self.thermal_boundaries:
          value = 0
          if thermal_boundary_1.id != thermal_boundary_2.id:
            value = thermal_boundary_2.opaque_area / (total_area - thermal_boundary_1.opaque_area)
          values.append(value)
        for thermal_boundary in self.thermal_boundaries:
          for thermal_opening in thermal_boundary.thermal_openings:
            value = thermal_opening.area / (total_area - thermal_boundary_1.opaque_area)
            values.append(value)
        view_factors_matrix.append(values)

      for thermal_boundary_1 in self.thermal_boundaries:
        values = []
        for thermal_opening_1 in thermal_boundary_1.thermal_openings:
          for thermal_boundary_2 in self.thermal_boundaries:
            value = thermal_boundary_2.opaque_area / (total_area - thermal_opening_1.area)
            values.append(value)
          for thermal_boundary in self.thermal_boundaries:
            for thermal_opening_2 in thermal_boundary.thermal_openings:
              value = 0
              if thermal_opening_1.id != thermal_opening_2.id:
                value = thermal_opening_2.area / (total_area - thermal_opening_1.area)
              values.append(value)
          view_factors_matrix.append(values)
      self._view_factors_matrix = view_factors_matrix
    return self._view_factors_matrix

  @property
  def usage_name(self) -> Union[None, str]:
    """
    Get thermal zone usage name
    :return: None or str
    """
    if self._usage_from_parent:
      if self._parent_internal_zone.usages is None:
        return None
      self._usage_name = ''
      for usage in self._parent_internal_zone.usages:
        self._usage_name += str(round(usage.percentage * 100)) + '-' + usage.name + '_'
      self._usage_name = self._usage_name[:-1]
    return self._usage_name

  @staticmethod
  def _get_schedule_of_day(requested_day_type, schedules):
    for schedule in schedules:
      for day_type in schedule.day_types:
        if day_type == requested_day_type:
          return schedule
      return None

  @property
  def hours_day(self) -> Union[None, float]:
    """
    Get thermal zone usage hours per day
    :return: None or float
    """
    if self.usages is None:
      return None
    if self._hours_day is None:
      self._hours_day = 0
      for usage in self.usages:
        self._hours_day += usage.percentage * usage.hours_day
    return self._hours_day

  @property
  def days_year(self) -> Union[None, float]:
    """
    Get thermal zone usage days per year
    :return: None or float
    """
    if self.usages is None:
      return None
    if self._days_year is None:
      self._days_year = 0
      for usage in self.usages:
        self._days_year += usage.percentage * usage.days_year
    return self._days_year

  @property
  def mechanical_air_change(self) -> Union[None, float]:
    """
    Get thermal zone mechanical air change in air change per second (1/s)
    :return: None or float
    """
    if self.usages is None:
      return None
    if self._mechanical_air_change is None:
      self._mechanical_air_change = 0
      for usage in self.usages:
        if usage.mechanical_air_change is None:
          return None
        self._mechanical_air_change += usage.percentage * usage.mechanical_air_change
    return self._mechanical_air_change

  @property
  def occupancy(self) -> Union[None, Occupancy]:
    """
    Get occupancy in the thermal zone
    :return: None or Occupancy
    """
    if self.usages is None:
      return None

    if self._occupancy is None:
      self._occupancy = Occupancy()
      _occupancy_density = 0
      _convective_part = 0
      _radiative_part = 0
      _latent_part = 0
      for usage in self.usages:
        if usage.occupancy is None:
          return None
        _occupancy_density += usage.percentage * usage.occupancy.occupancy_density
        if usage.occupancy.sensible_convective_internal_gain is not None:
          _convective_part += usage.percentage * usage.occupancy.sensible_convective_internal_gain
          _radiative_part += usage.percentage * usage.occupancy.sensible_radiative_internal_gain
          _latent_part += usage.percentage * usage.occupancy.latent_internal_gain
      self._occupancy.occupancy_density = _occupancy_density
      self._occupancy.sensible_convective_internal_gain = _convective_part
      self._occupancy.sensible_radiative_internal_gain = _radiative_part
      self._occupancy.latent_internal_gain = _latent_part

      _occupancy_reference = self.usages[0].occupancy
      if _occupancy_reference.occupancy_schedules is not None:
        _schedules = []
        for schedule_index, schedule_value in enumerate(_occupancy_reference.occupancy_schedules):
          schedule = Schedule()
          schedule.type = schedule_value.type
          schedule.day_types = schedule_value.day_types
          schedule.data_type = schedule_value.data_type
          schedule.time_step = schedule_value.time_step
          schedule.time_range = schedule_value.time_range

          new_values = []
          for i_value, _ in enumerate(schedule_value.values):
            _new_value = 0
            for usage in self.usages:
              _new_value += usage.percentage * usage.occupancy.occupancy_schedules[schedule_index].values[i_value]
            new_values.append(_new_value)
          schedule.values = new_values
          _schedules.append(schedule)
        self._occupancy.occupancy_schedules = _schedules
    return self._occupancy

  @property
  def lighting(self) -> Union[None, Lighting]:
    """
    Get lighting information
    :return: None or Lighting
    """
    if self.usages is None:
      return None

    if self._lighting is None:
      self._lighting = Lighting()
      _lighting_density = 0
      _convective_part = 0
      _radiative_part = 0
      _latent_part = 0
      for usage in self.usages:
        if usage.lighting is None:
          return None
        _lighting_density += usage.percentage * usage.lighting.density
        if usage.lighting.convective_fraction is not None:
          _convective_part += (
              usage.percentage * usage.lighting.density * usage.lighting.convective_fraction
          )
          _radiative_part += (
              usage.percentage * usage.lighting.density * usage.lighting.radiative_fraction
          )
          _latent_part += (
              usage.percentage * usage.lighting.density * usage.lighting.latent_fraction
          )
      self._lighting.density = _lighting_density
      if _lighting_density > 0:
        self._lighting.convective_fraction = _convective_part / _lighting_density
        self._lighting.radiative_fraction = _radiative_part / _lighting_density
        self._lighting.latent_fraction = _latent_part / _lighting_density
      else:
        self._lighting.convective_fraction = 0
        self._lighting.radiative_fraction = 0
        self._lighting.latent_fraction = 0

      _lighting_reference = self.usages[0].lighting
      if _lighting_reference.schedules is not None:
        _schedules = []
        for schedule_index, schedule_value in enumerate(_lighting_reference.schedules):
          schedule = Schedule()
          schedule.type = schedule_value.type
          schedule.day_types = schedule_value.day_types
          schedule.data_type = schedule_value.data_type
          schedule.time_step = schedule_value.time_step
          schedule.time_range = schedule_value.time_range

          new_values = []
          for i_value, _ in enumerate(schedule_value.values):
            _new_value = 0
            for usage in self.usages:
              _new_value += usage.percentage * usage.lighting.schedules[schedule_index].values[i_value]
            new_values.append(_new_value)
          schedule.values = new_values
          _schedules.append(schedule)
        self._lighting.schedules = _schedules
    return self._lighting

  @property
  def appliances(self) -> Union[None, Appliances]:
    """
    Get appliances information
    :return: None or Appliances
    """
    if self.usages is None:
      return None

    if self._appliances is None:
      self._appliances = Appliances()
      _appliances_density = 0
      _convective_part = 0
      _radiative_part = 0
      _latent_part = 0
      for usage in self.usages:
        if usage.appliances is None:
          return None
        _appliances_density += usage.percentage * usage.appliances.density
        if usage.appliances.convective_fraction is not None:
          _convective_part += (
              usage.percentage * usage.appliances.density * usage.appliances.convective_fraction
          )
          _radiative_part += (
              usage.percentage * usage.appliances.density * usage.appliances.radiative_fraction
          )
          _latent_part += (
              usage.percentage * usage.appliances.density * usage.appliances.latent_fraction
          )
      self._appliances.density = _appliances_density
      if _appliances_density > 0:
        self._appliances.convective_fraction = _convective_part / _appliances_density
        self._appliances.radiative_fraction = _radiative_part / _appliances_density
        self._appliances.latent_fraction = _latent_part / _appliances_density
      else:
        self._appliances.convective_fraction = 0
        self._appliances.radiative_fraction = 0
        self._appliances.latent_fraction = 0

      _appliances_reference = self.usages[0].appliances
      if _appliances_reference.schedules is not None:
        _schedules = []
        for schedule_index, schedule_value in enumerate(_appliances_reference.schedules):
          schedule = Schedule()
          schedule.type = schedule_value.type
          schedule.day_types = schedule_value.day_types
          schedule.data_type = schedule_value.data_type
          schedule.time_step = schedule_value.time_step
          schedule.time_range = schedule_value.time_range

          new_values = []
          for i_value, _ in enumerate(schedule_value.values):
            _new_value = 0
            for usage in self.usages:
              _new_value += usage.percentage * usage.appliances.schedules[schedule_index].values[i_value]
            new_values.append(_new_value)
          schedule.values = new_values
          _schedules.append(schedule)
        self._appliances.schedules = _schedules
    return self._appliances

  @property
  def internal_gains(self) -> Union[None, List[InternalGain]]:
    """
    Calculates and returns the list of all internal gains defined
    :return: [InternalGain]
    """
    if self.usages is None:
      return None

    if self._internal_gains is None:
      _internal_gain = InternalGain()
      _days = [cte.MONDAY, cte.TUESDAY, cte.WEDNESDAY, cte.THURSDAY, cte.FRIDAY, cte.SATURDAY, cte.SUNDAY, cte.HOLIDAY]
      _average_internal_gain = 0
      _convective_fraction = 0
      _radiative_fraction = 0
      _latent_fraction = 0
      _schedules = None
      _base_schedule = Schedule()
      _base_schedule.type = cte.INTERNAL_GAINS
      _base_schedule.time_range = cte.DAY
      _base_schedule.time_step = cte.HOUR
      _base_schedule.data_type = cte.ANY_NUMBER
      _schedules_defined = True
      values = numpy.zeros([24, 8])
      for usage in self.usages:
        for internal_gain in usage.internal_gains:
          _average_internal_gain += internal_gain.average_internal_gain * usage.percentage
          _convective_fraction += (
              internal_gain.average_internal_gain * usage.percentage * internal_gain.convective_fraction
          )
          _radiative_fraction += (
              internal_gain.average_internal_gain * usage.percentage * internal_gain.radiative_fraction
          )
          _latent_fraction += (
            internal_gain.average_internal_gain * usage.percentage * internal_gain.latent_fraction
          )
      for usage in self.usages:
        for internal_gain in usage.internal_gains:
          if internal_gain.schedules is None:
            _schedules_defined = False
            break
          if len(internal_gain.schedules) == 0:
            _schedules_defined = False
            break
          for day, _schedule in enumerate(internal_gain.schedules):
            for v_index, value in enumerate(_schedule.values):
              values[v_index, day] += value * usage.percentage

      if _schedules_defined:
        _schedules = []
        for day_index, day in enumerate(_days):
          _schedule = copy.deepcopy(_base_schedule)
          _schedule.day_types = [day]
          _schedule.values = values[:day_index]
          _schedules.append(_schedule)

      _internal_gain.average_internal_gain = _average_internal_gain
      _internal_gain.convective_fraction = 0
      _internal_gain.radiative_fraction = 0
      _internal_gain.latent_fraction = 0
      if _average_internal_gain != 0:
        _internal_gain.convective_fraction = _convective_fraction / _average_internal_gain
        _internal_gain.radiative_fraction = _radiative_fraction / _average_internal_gain
        _internal_gain.latent_fraction = _latent_fraction / _average_internal_gain
      _internal_gain.type = 'mean_value'
      _internal_gain.schedules = _schedules
      self._internal_gains = [_internal_gain]
    return self._internal_gains

  @property
  def thermal_control(self) -> Union[None, ThermalControl]:
    """
    Get thermal control of this thermal zone
    :return: None or ThermalControl
    """
    if self.usages is None:
      return None

    if self._thermal_control is not None:
      return self._thermal_control
    self._thermal_control = ThermalControl()
    _mean_heating_set_point = 0
    _heating_set_back = 0
    _mean_cooling_set_point = 0
    for usage in self.usages:
      _mean_heating_set_point += usage.percentage * usage.thermal_control.mean_heating_set_point
      _heating_set_back += usage.percentage * usage.thermal_control.heating_set_back
      _mean_cooling_set_point += usage.percentage * usage.thermal_control.mean_cooling_set_point
    self._thermal_control.mean_heating_set_point = _mean_heating_set_point
    self._thermal_control.heating_set_back = _heating_set_back
    self._thermal_control.mean_cooling_set_point = _mean_cooling_set_point

    _thermal_control_reference = self.usages[0].thermal_control
    _types_reference = []
    if _thermal_control_reference.hvac_availability_schedules is not None:
      _types_reference.append([cte.HVAC_AVAILABILITY, _thermal_control_reference.hvac_availability_schedules])
    if _thermal_control_reference.heating_set_point_schedules is not None:
      _types_reference.append([cte.HEATING_SET_POINT, _thermal_control_reference.heating_set_point_schedules])
    if _thermal_control_reference.cooling_set_point_schedules is not None:
      _types_reference.append([cte.COOLING_SET_POINT, _thermal_control_reference.cooling_set_point_schedules])

    for i_type, _ in enumerate(_types_reference):
      _schedules = []
      _schedule_type = _types_reference[i_type][1]
      for i_schedule, schedule_value in enumerate(_schedule_type):
        schedule = Schedule()
        schedule.type = schedule_value.type
        schedule.day_types = schedule_value.day_types
        schedule.data_type = schedule_value.data_type
        schedule.time_step = schedule_value.time_step
        schedule.time_range = schedule_value.time_range

        new_values = []
        for i_value, _ in enumerate(schedule_value.values):
          _new_value = 0
          for usage in self.usages:
            if _types_reference[i_type][0] == cte.HVAC_AVAILABILITY:
              _new_value += usage.percentage * \
                            usage.thermal_control.hvac_availability_schedules[i_schedule].values[i_value]
            elif _types_reference[i_type][0] == cte.HEATING_SET_POINT:
              _new_value += usage.percentage * \
                            usage.thermal_control.heating_set_point_schedules[i_schedule].values[i_value]
            elif _types_reference[i_type][0] == cte.COOLING_SET_POINT:
              _new_value += usage.percentage * \
                            usage.thermal_control.cooling_set_point_schedules[i_schedule].values[i_value]
          new_values.append(_new_value)
        schedule.values = new_values
        _schedules.append(schedule)
      if i_type == 0:
        self._thermal_control.hvac_availability_schedules = _schedules
      elif i_type == 1:
        self._thermal_control.heating_set_point_schedules = _schedules
      elif i_type == 2:
        self._thermal_control.cooling_set_point_schedules = _schedules
    return self._thermal_control

  @property
  def domestic_hot_water(self) -> Union[None, DomesticHotWater]:
    """
    Get domestic hot water information of this thermal zone
    :return: None or DomesticHotWater
    """
    self._domestic_hot_water = DomesticHotWater()
    _mean_peak_density_load = 0
    _mean_peak_flow = 0
    _mean_service_temperature = 0
    for usage in self.usages:
      if usage.domestic_hot_water.density is not None:
        _mean_peak_density_load += usage.percentage * usage.domestic_hot_water.density
      if usage.domestic_hot_water.peak_flow is not None:
        _mean_peak_flow += usage.percentage * usage.domestic_hot_water.peak_flow
      if usage.domestic_hot_water.service_temperature is not None:
        _mean_service_temperature += usage.percentage * usage.domestic_hot_water.service_temperature
    self._domestic_hot_water.density = _mean_peak_density_load
    self._domestic_hot_water.peak_flow = _mean_peak_flow
    self._domestic_hot_water.service_temperature = _mean_service_temperature

    _domestic_hot_water_reference = self.usages[0].domestic_hot_water
    if _domestic_hot_water_reference.schedules is not None:
      _schedules = []
      for schedule_index, schedule_value in enumerate(_domestic_hot_water_reference.schedules):
        schedule = Schedule()
        schedule.type = schedule_value.type
        schedule.day_types = schedule_value.day_types
        schedule.data_type = schedule_value.data_type
        schedule.time_step = schedule_value.time_step
        schedule.time_range = schedule_value.time_range

        new_values = []
        for i_value, _ in enumerate(schedule_value.values):
          _new_value = 0
          for usage in self.usages:
            _new_value += usage.percentage * usage.domestic_hot_water.schedules[schedule_index].values[i_value]
          new_values.append(_new_value)
        schedule.values = new_values
        _schedules.append(schedule)
      self._domestic_hot_water.schedules = _schedules

    return self._domestic_hot_water

  @property
  def total_floor_area(self):
    """
    Get the total floor area of this thermal zone in m2
    :return: float
    """
    self._total_floor_area = self.footprint_area * self._number_of_storeys
    return self._total_floor_area



"""
TestSystemsCatalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from unittest import TestCase
from hub.catalog_factories.energy_systems_catalog_factory import EnergySystemsCatalogFactory


class TestSystemsCatalog(TestCase):

  def test_montreal_custom_catalog(self):
    catalog = EnergySystemsCatalogFactory('montreal_custom').catalog
    catalog_categories = catalog.names()
    archetypes = catalog.names('archetypes')
    self.assertEqual(23, len(archetypes['archetypes']))
    systems = catalog.names('systems')
    self.assertEqual(18, len(systems['systems']))
    generation_equipments = catalog.names('generation_equipments')
    self.assertEqual(7, len(generation_equipments['generation_equipments']))
    distribution_equipments = catalog.names('distribution_equipments')
    self.assertEqual(8, len(distribution_equipments['distribution_equipments']))
    emission_equipments = catalog.names('emission_equipments')
    self.assertEqual(3, len(emission_equipments['emission_equipments']))
    with self.assertRaises(ValueError):
      catalog.names('unknown')

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')



"""
Thermal zones creation module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.imports.construction.helpers.storeys_generation import StoreysGeneration


class ThermalZonesCreation:
  """
  PeakLoads class
  """
  def __init__(self, building=None):
    self._building = building

#  # The agreement is that the layers are defined from outside to inside
#  external_layer = catalog_construction.layers[0]
#  external_surface = thermal_boundary.parent_surface
#  external_surface.short_wave_reflectance = 1 - external_layer.material.solar_absorptance
#  external_surface.long_wave_emittance = 1 - external_layer.material.solar_absorptance
#  internal_layer = catalog_construction.layers[len(catalog_construction.layers) - 1]
#  internal_surface = thermal_boundary.internal_surface
#  internal_surface.short_wave_reflectance = 1 - internal_layer.material.solar_absorptance
#  internal_surface.long_wave_emittance = 1 - internal_layer.material.solar_absorptance

  @property
  def thermal_zones_from_storeys(self):
    """
    Create and get thermal zones as 1 per each storey
    :return: [ThermalZone]
    """
    raise NotImplementedError

  @staticmethod
  def _create_storeys(building, archetype, divide_in_storeys):
    building.average_storey_height = archetype.average_storey_height
    thermal_zones = StoreysGeneration(building, building.internal_zones[0],
                                      divide_in_storeys=divide_in_storeys).thermal_zones
    building.internal_zones[0].thermal_zones_from_internal_zones = thermal_zones



"""
Surface module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations

import math
import uuid
from typing import List, Union
import numpy as np
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.attributes.plane import Plane
from hub.city_model_structure.attributes.point import Point
from hub.city_model_structure.greenery.vegetation import Vegetation
from hub.city_model_structure.building_demand.thermal_boundary import ThermalBoundary
import hub.helpers.constants as cte


class Surface:
  """
  Surface class
  """

  def __init__(self, solid_polygon, perimeter_polygon, holes_polygons=None, name=None, surface_type=None):
    self._type = surface_type
    self._name = name
    self._id = None
    self._azimuth = None
    self._inclination = None
    self._area = None
    self._lower_corner = None
    self._upper_corner = None
    self._global_irradiance = {}
    self._perimeter_polygon = perimeter_polygon
    self._holes_polygons = holes_polygons
    self._solid_polygon = solid_polygon
    self._short_wave_reflectance = None
    self._long_wave_emittance = None
    self._inverse = None
    self._associated_thermal_boundaries = []
    self._vegetation = None
    self._percentage_shared = None
    self._solar_collectors_area_reduction_factor = None

  @property
  def name(self):
    """
    Get the surface name
    :return: str
    """
    if self._name is None:
      self._name = str(uuid.uuid4())
    return self._name

  @property
  def id(self):
    """
    Get the surface id
    :return: str
    """
    if self._id is None:
      raise ValueError('Undefined surface id')
    return self._id

  @id.setter
  def id(self, value):
    """
    Set the surface id
    :param value: str
    """
    if value is not None:
      self._id = str(value)

  def _max_coord(self, axis):
    if axis == 'x':
      axis = 0
    elif axis == 'y':
      axis = 1
    else:
      axis = 2
    max_coordinate = ''
    for point in self.perimeter_polygon.coordinates:
      if max_coordinate == '':
        max_coordinate = point[axis]
      elif max_coordinate < point[axis]:
        max_coordinate = point[axis]
    return max_coordinate

  def _min_coord(self, axis):
    if axis == 'x':
      axis = 0
    elif axis == 'y':
      axis = 1
    else:
      axis = 2
    min_coordinate = ''
    for point in self.perimeter_polygon.coordinates:
      if min_coordinate == '':
        min_coordinate = point[axis]
      elif min_coordinate > point[axis]:
        min_coordinate = point[axis]
    return min_coordinate

  @property
  def lower_corner(self):
    """
    Get surface's lower corner [x, y, z]
    :return: [float]
    """
    if self._lower_corner is None:
      self._lower_corner = [self._min_coord('x'), self._min_coord('y'), self._min_coord('z')]
    return self._lower_corner

  @property
  def upper_corner(self):
    """
    Get surface's upper corner [x, y, z]
    :return: [float]
    """
    if self._upper_corner is None:
      self._upper_corner = [self._max_coord('x'), self._max_coord('y'), self._max_coord('z')]
    return self._upper_corner

  @property
  def perimeter_area(self):
    """
    Get perimeter surface area in square meters (opaque + transparent)
    :return: float
    """
    self._area = self.perimeter_polygon.area
    return self._area

  @property
  def azimuth(self):
    """
    Get surface azimuth in radians (north = 0)
    :return: float
    """
    if self._azimuth is None:
      normal = self.perimeter_polygon.normal
      self._azimuth = np.arctan2(normal[1], normal[0])
    return self._azimuth

  @property
  def inclination(self):
    """
    Get surface inclination in radians (zenith = 0, horizon = pi/2)
    :return: float
    """
    if self._inclination is None:
      self._inclination = np.arccos(self.perimeter_polygon.normal[2])
    return self._inclination

  @property
  def type(self):
    """
    Get surface type Ground, Ground wall, Wall, Attic floor, Interior slab, Interior wall, Roof or Virtual internal
    If the geometrical LoD is lower than 4,
    the surfaces' types are not defined in the importer and can only be Ground, Wall or Roof
    :return: str
    """
    if self._type is None:
      inclination_cos = math.cos(self.inclination)
      # 170 degrees
      if inclination_cos <= -0.98:
        self._type = 'Ground'
      # between 80 and 100 degrees
      elif abs(inclination_cos) <= 0.17:
        self._type = 'Wall'
      else:
        self._type = 'Roof'
    return self._type

  @property
  def global_irradiance(self) -> dict:
    """
    Get global irradiance on surface in J/m2
    :return: dict
    """
    return self._global_irradiance

  @global_irradiance.setter
  def global_irradiance(self, value):
    """
    Set global irradiance on surface in J/m2
    :param value: dict
    """
    self._global_irradiance = value

  @property
  def perimeter_polygon(self) -> Polygon:
    """
    Get a polygon surface defined by the perimeter, merging solid and holes
    :return: Polygon
    """
    return self._perimeter_polygon

  @property
  def solid_polygon(self) -> Polygon:
    """
    Get the solid surface
    :return: Polygon
    """
    return self._solid_polygon

  @solid_polygon.setter
  def solid_polygon(self, value):
    """
    Set the solid surface
    :return: Polygon
    """
    self._solid_polygon = value

  @property
  def holes_polygons(self) -> Union[List[Polygon], None]:
    """
    Get hole surfaces, a list of hole polygons found in the surface
    :return: None, [] or [Polygon]
    None -> not known whether holes exist in reality or not due to low level of detail of input data
    [] -> no holes in the surface
    [Polygon] -> one or more holes in the surface
    """
    return self._holes_polygons

  @holes_polygons.setter
  def holes_polygons(self, value):
    """
    Set the hole surfaces
    :param value: [Polygon]
    """
    self._holes_polygons = value

  @property
  def short_wave_reflectance(self):
    """
    Get the short wave reflectance, this includes all solar spectrum, visible and not visible
    The absorptance as an opaque surface, can be calculated as 1-short_wave_reflectance
    :return: float
    """
    return self._short_wave_reflectance

  @short_wave_reflectance.setter
  def short_wave_reflectance(self, value):
    """
    Set the short wave reflectance, this includes all solar spectrum, visible and not visible
    The absorptance as an opaque surface, can be calculated as 1-short_wave_reflectance
    :param value: float
    """
    self._short_wave_reflectance = value

  @property
  def long_wave_emittance(self):
    """
    Get the long wave emittance af the surface
    The thermal absorptance can be calculated as 1-long_wave_emittance
    :return: float
    """
    return self._long_wave_emittance

  @long_wave_emittance.setter
  def long_wave_emittance(self, value):
    """
    Set the long wave emittance af the surface
    The thermal absorptance can be calculated as 1-long_wave_emittance
    :param value: float
    """
    self._long_wave_emittance = value

  @property
  def inverse(self) -> Surface:
    """
    Get the inverse surface (the same surface pointing backwards)
    :return: Surface
    """
    if self._inverse is None:
      new_solid_polygon = Polygon(self.solid_polygon.inverse)
      new_perimeter_polygon = Polygon(self.perimeter_polygon.inverse)
      new_holes_polygons = []
      if self.holes_polygons is not None:
        for hole in self.holes_polygons:
          new_holes_polygons.append(Polygon(hole.inverse))
      else:
        new_holes_polygons = None
      self._inverse = Surface(new_solid_polygon, new_perimeter_polygon, new_holes_polygons, cte.VIRTUAL_INTERNAL)
    return self._inverse

  def divide(self, z):
    """
    Divides a surface at Z plane
    :return: Surface, Surface, Any
    """
    # todo: check return types
    #  recheck this method for LoD3 (windows)
    origin = Point([0, 0, z])
    normal = np.array([0, 0, 1])
    plane = Plane(normal=normal, origin=origin)
    polygon = self.perimeter_polygon
    part_1, part_2, intersection = polygon.divide(plane)
    surface_child = Surface(part_1, part_1, name=self.name, surface_type=self.type)
    rest_surface = Surface(part_2, part_2, name=self.name, surface_type=self.type)
    return surface_child, rest_surface, intersection

  @property
  def vegetation(self) -> Union[None, Vegetation]:
    """
    Get the vegetation construction at the external surface of the thermal boundary
    :return: None or Vegetation
    """
    return self._vegetation

  @vegetation.setter
  def vegetation(self, value):
    """
    Set the vegetation construction at the external surface of the thermal boundary
    :param value: Vegetation
    """
    self._vegetation = value

  @property
  def associated_thermal_boundaries(self) -> Union[None, List[ThermalBoundary]]:
    """
    Get the list of thermal boundaries that has this surface as external face
    :return: None or [ThermalBoundary]
    """
    return self._associated_thermal_boundaries

  @associated_thermal_boundaries.setter
  def associated_thermal_boundaries(self, value):
    """
    Set the list of thermal boundaries that has this surface as external face
    :param value: None or [ThermalBoundary]
    """
    self._associated_thermal_boundaries = value

  @property
  def percentage_shared(self):
    """
    Get percentage of the wall shared with other walls
    :return: float
    """
    return self._percentage_shared

  @percentage_shared.setter
  def percentage_shared(self, value):
    """
    Set percentage of the wall shared with other walls
    :param value: float
    """
    self._percentage_shared = value

  @property
  def solar_collectors_area_reduction_factor(self):
    """
    Get factor area collector per surface area if set or calculate using Romero Rodriguez, L. et al (2017) model if not
    :return: float
    """
    if self._solar_collectors_area_reduction_factor is None:
      if self.type == cte.ROOF:
        _protected_building_restriction = 1
        # 10 degrees range
        if abs(math.sin(self.inclination)) < 0.17:
          # horizontal
          _construction_restriction = 0.8
          _separation_of_panels = 0.46
          _shadow_between_panels = 0.7
        else:
          # tilted
          _construction_restriction = 0.9
          _separation_of_panels = 0.9
          _shadow_between_panels = 1
        self._solar_collectors_area_reduction_factor = (
            _protected_building_restriction * _construction_restriction * _separation_of_panels * _shadow_between_panels
        )
    return self._solar_collectors_area_reduction_factor

  @solar_collectors_area_reduction_factor.setter
  def solar_collectors_area_reduction_factor(self, value):
    """
    Set factor area collector per surface area
    :param value: float
    """
    self._solar_collectors_area_reduction_factor = value



"""
Construction catalog material
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class Material:
  """
  Material class
  """
  def __init__(self, material_id,
               name,
               solar_absorptance,
               thermal_absorptance,
               visible_absorptance,
               no_mass=False,
               thermal_resistance=None,
               conductivity=None,
               density=None,
               specific_heat=None):
    self._id = material_id
    self._name = name
    self._solar_absorptance = solar_absorptance
    self._thermal_absorptance = thermal_absorptance
    self._visible_absorptance = visible_absorptance
    self._no_mass = no_mass
    self._thermal_resistance = thermal_resistance
    self._conductivity = conductivity
    self._density = density
    self._specific_heat = specific_heat

  @property
  def id(self):
    """
    Get material id
    :return: str
    """
    return self._id

  @property
  def name(self):
    """
    Get material name
    :return: str
    """
    return self._name

  @property
  def conductivity(self):
    """
    Get material conductivity in W/mK
    :return: None or float
    """
    return self._conductivity

  @property
  def specific_heat(self):
    """
    Get material conductivity in J/kgK
    :return: None or float
    """
    return self._specific_heat

  @property
  def density(self):
    """
    Get material density in kg/m3
    :return: None or float
    """
    return self._density

  @property
  def solar_absorptance(self):
    """
    Get material solar absorptance
    :return: None or float
    """
    return self._solar_absorptance

  @property
  def thermal_absorptance(self):
    """
    Get material thermal absorptance
    :return: None or float
    """
    return self._thermal_absorptance

  @property
  def visible_absorptance(self):
    """
    Get material visible absorptance
    :return: None or float
    """
    return self._visible_absorptance

  @property
  def no_mass(self):
    """
    Get material no mass flag
    :return: None or Boolean
    """
    return self._no_mass

  @property
  def thermal_resistance(self):
    """
    Get material thermal resistance in m2K/W
    :return: None or float
    """
    return self._thermal_resistance

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Material': {'id': self.id,
                            'name': self.name,
                            'is no-mass': self.no_mass,
                            'density [kg/m3]': self.density,
                            'specific heat [J/kgK]': self.specific_heat,
                            'conductivity [W/mK]': self.conductivity,
                            'thermal resistance [m2K/W]': self.thermal_resistance,
                            'solar absorptance': self.solar_absorptance,
                            'thermal absorptance': self.thermal_absorptance,
                            'visible absorptance': self.visible_absorptance
                            }
               }
    return content



"""
TestGeometryFactory test and validate the city model structure geometric parameters
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.helpers.dictionaries import Dictionaries
from hub.imports.usage_factory import UsageFactory
from hub.imports.construction_factory import ConstructionFactory


class TestGeometryFactory(TestCase):
  """
  Non-functional TestGeometryFactory
  Load testing
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._city = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()

  def _get_citygml(self, file):
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('citygml', path=file_path).city
    self.assertIsNotNone(self._city, 'city is none')
    return self._city

  def _check_result(self, city):
    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self.assertIsNot(len(internal_zone.usages), 0, 'no building usages defined')
        for usage in internal_zone.usages:
          self.assertIsNotNone(usage.id, 'usage id is none')
        for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
          self._check_thermal_zone(thermal_zone)

  def _check_buildings(self, city):
    for building in city.buildings:
      self.assertIsNotNone(building.internal_zones, 'no internal zones created')
      for internal_zone in building.internal_zones:
        self.assertIsNotNone(internal_zone.usages, 'usage zones are not defined')
        self.assertIsNotNone(internal_zone.thermal_zones_from_internal_zones, 'thermal zones are not defined')
      self.assertIsNone(building.basement_heated, 'building basement_heated is not none')
      self.assertIsNone(building.attic_heated, 'building attic_heated is not none')
      self.assertIsNotNone(building.average_storey_height, 'building average_storey_height is none')
      self.assertIsNotNone(building.storeys_above_ground, 'building storeys_above_ground is none')
      self.assertTrue(building.is_conditioned, 'building is_conditioned is not conditioned')

  def _check_thermal_zone(self, thermal_zone):
    self.assertIsNotNone(thermal_zone.id, 'thermal_zone id is none')
    self.assertIsNotNone(thermal_zone.usage_name, 'thermal_zone usage is not none')
    self.assertIsNotNone(thermal_zone.hours_day, 'thermal_zone hours a day is none')
    self.assertIsNotNone(thermal_zone.days_year, 'thermal_zone days a year is none')
    self.assertIsNotNone(thermal_zone.occupancy, 'thermal_zone occupancy is none')
    self.assertIsNotNone(thermal_zone.thermal_control, 'thermal_zone thermal control is none')
    self.assertIsNotNone(thermal_zone.internal_gains, 'thermal_zone internal gains returns none')

  def _check_extra_thermal_zone(self, thermal_zone):
    self.assertIsNotNone(thermal_zone.lighting, 'thermal_zone lighting is none')
    self.assertIsNotNone(thermal_zone.appliances, 'thermal_zone appliances is none')
    self.assertIsNotNone(thermal_zone.mechanical_air_change, 'thermal_zone mechanical air change is none')

  @staticmethod
  def _prepare_case_usage_first(city, input_key, construction_key, usage_key):
    if input_key == 'pluto':
      for building in city.buildings:
        building.function = Dictionaries().pluto_function_to_hub_function[building.function]
    elif input_key == 'hft':
      for building in city.buildings:
        building.function = Dictionaries().hft_function_to_hub_function[building.function]
    UsageFactory(usage_key, city).enrich()
    ConstructionFactory(construction_key, city).enrich()

  @staticmethod
  def _prepare_case_construction_first(city, input_key, construction_key, usage_key):
    if input_key == 'pluto':
      for building in city.buildings:
        building.function = Dictionaries().pluto_function_to_hub_function[building.function]
    elif input_key == 'hft':
      for building in city.buildings:
        building.function = Dictionaries().hft_function_to_hub_function[building.function]
    ConstructionFactory(construction_key, city).enrich()
    UsageFactory(usage_key, city).enrich()

  def _test_hft(self, file):
    _construction_keys = ['nrel']
    _usage_keys = ['comnet']
    for construction_key in _construction_keys:
      for usage_key in _usage_keys:
        print(construction_key, usage_key)
        # construction factory called first
        city = self._get_citygml(file)
        for building in city.buildings:
          building.year_of_construction = 2006
        self.assertTrue(len(city.buildings) > 0)
        self._prepare_case_construction_first(city, 'hft', construction_key, usage_key)
        self._check_result(city)
        if usage_key == 'comnet':
          for building in city.buildings:
            for internal_zone in building.internal_zones:
              for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
                self._check_extra_thermal_zone(thermal_zone)
        # usage factory called first
        city = self._get_citygml(file)
        for building in city.buildings:
          building.year_of_construction = 2006
        self.assertTrue(len(city.buildings) > 0)
        self._prepare_case_usage_first(city, 'hft', construction_key, usage_key)
        self._check_result(city)
        if usage_key == 'comnet':
          for building in city.buildings:
            for internal_zone in building.internal_zones:
              for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
                self._check_extra_thermal_zone(thermal_zone)

  def _test_pluto(self, file):
    _construction_keys = ['nrel']
    _usage_keys = ['comnet', 'nrcan']
    for construction_key in _construction_keys:
      for usage_key in _usage_keys:
        # construction factory called first
        city = self._get_citygml(file)
        for building in city.buildings:
          building.year_of_construction = 2006
        self.assertTrue(len(city.buildings) > 0)
        self._prepare_case_construction_first(city, 'pluto', construction_key, usage_key)
        self._check_result(city)
        if usage_key == 'comnet':
          for building in city.buildings:
            for internal_zone in building.internal_zones:
              for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
                self._check_extra_thermal_zone(thermal_zone)
        # usage factory called first
        city = self._get_citygml(file)
        for building in city.buildings:
          building.year_of_construction = 2006
        self.assertTrue(len(city.buildings) > 0)
        self._prepare_case_usage_first(city, 'pluto', construction_key, usage_key)
        self._check_result(city)
        if usage_key == 'comnet':
          for building in city.buildings:
            for internal_zone in building.internal_zones:
              for thermal_zone in internal_zone.thermal_zones_from_internal_zones:
                self._check_extra_thermal_zone(thermal_zone)

  def test_enrichment(self):
    """
    Test enrichment of the city with different orders
    :return: None
    """
    file_1 = 'one_building_in_kelowna.gml'
    self._test_hft(file_1)
    file_2 = 'C40_Final.gml'
    self._test_hft(file_2)



"""
Energy System catalog content
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class Content:
  """
  Content class
  """
  def __init__(self, archetypes, systems, generations, distributions, emissions):
    self._archetypes = archetypes
    self._systems = systems
    self._generations = generations
    self._distributions = distributions
    self._emissions = emissions

  @property
  def archetypes(self):
    """
    All archetype system clusters in the catalog
    """
    return self._archetypes

  @property
  def systems(self):
    """
    All systems in the catalog
    """
    return self._systems

  @property
  def generation_equipments(self):
    """
    All generation equipments in the catalog
    """
    return self._generations

  @property
  def distribution_equipments(self):
    """
    All distribution equipments in the catalog
    """
    return self._distributions

  @property
  def emission_equipments(self):
    """
    All emission equipments in the catalog
    """
    return self._emissions

  def to_dictionary(self):
    """Class content to dictionary"""
    _archetypes = []
    for _archetype in self.archetypes:
      _archetypes.append(_archetype.to_dictionary())
    content = {'Archetypes': _archetypes}

    return content

  def __str__(self):
    """Print content"""
    _archetypes = []
    for _archetype in self.archetypes:
      _archetypes.append(_archetype.to_dictionary())
    content = {'Archetypes': _archetypes}

    return str(content)



"""
HeatPumpExport exports heatpump outputs into several files after insel execution
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import os
import logging
import subprocess
from typing import List, Union, Dict
import yaml
from string import Template
import pandas as pd


class HeatPumpExport:
  """
  Exports heat pump values as coefficients
  of some defined function
  """

  def __init__(self, base_path, city, output_path, template, demand_path=None, water_temp=None):
    self._template_path = template
    self._water_temp = water_temp
    self._constants_path = (base_path / 'heat_pumps/constants.yaml')
    # needed to compute max demand.
    self._demand_path = (base_path / 'heat_pumps/demand.txt') if demand_path is None else demand_path
    self._city = city
    self._input_data = None
    self._base_path = base_path
    self._output_path = output_path

  def _run_insel(self, user_input: Dict, capacity_coefficient: List, filename: str) -> Union[Dict, None]:
    """
    Runs insel and write the necessary files
    :param user_input: a dictionary containing the user
    values necessary to run insel
    :param capacity_coefficient: a list containing capacity coefficients
    :param filename: the name of the insel file to be created
    :return:
    """
    self._input_data = user_input
    self._update_input_data_with_coff(capacity_coefficient)
    # update input data with constants
    self._update_input_data_with_constants()
    # update input data with input and output files for insel
    self._update_input_data_with_files()
    insel_file_handler = None
    insel_template_handler = None
    try:
      # run insel
      insel_template_handler = open(self._template_path, "r")
      insel_template_content = insel_template_handler.read()
      insel_template = Template(insel_template_content).substitute(self._input_data)
      # create the insel file and write the template with substituted values into it
      insel_file = (self._base_path / 'heat_pumps' / filename)
      insel_file_handler = open(insel_file, "w")
      insel_file_handler.write(insel_template)
      # Now run insel
      self._delete_existing_output_files()
      logging.info(f'Running Insel with user input: {user_input} and coefficients {capacity_coefficient}')
      subprocess.run(['insel', str(insel_file)])
      # Writer headers to csv output files generated by insel
      self._write_insel_output_headers()
      # User output
      return self._get_user_out_put()
    except IOError as err:
      logging.error(f'An I/O error occurred while running insel: {str(err)}')
    finally:
      insel_file_handler.close()
      insel_template_handler.close()

  def _write_insel_output_headers(self):
    """
    Write headers to the various csv file generated by insel
    :return:
    """
    header = [
      'Year', ' Month', ' Day', 'Hour', 'Minute', 'HP Heat Output (kW)', 'Heating Demand (kW)', 'HP output flow rate',
      'Building Required Flow Rate', 'TES Charging Rate (kg/s)', 'Water Flow Rate After Splitter',
      'water temperature after splitter', 'TES Discharging Rate (kg/s)', 'TES discharge temperature',
      'Mixer Outlet Flow Rate (kg/s)', 'Mixer outlet temperature', 'Auxiliary heater fuel flow rate',
      'Auxiliary heater energy input (kW)', 'Building Inlet Flow Rate (kg/s)', 'Building inlet temperature',
      'Building return temperature', 'TES Return Flow Rate (kg/s)', 'TES return temperature',
      'TES Bypass Line Flow Rate (kg/s)', 'TES bypass line temperature', 'Flow Rate from TES to mixer 2 (kg/s)',
      'Temperature from Tes to mixer', 'HP Inlet Flow Rate (kg/s)', 'HP Inlet temperature', 'TES Node 1 Temperature',
      'TES Node 2 Temperature', 'TES Node 3 Temperature', 'TES Node 4 Temperature', 'TES Energy Content (J)',
      'HP Electricity Consumption (kW)', 'HP COP', 'Ambient Temperature', 'HP Operational Cost (CAD)',
      'Auxiliary Heater Operational Cost (CAD)', 'Operational CO2 Emissions of HP (g)',
      'Operational CO2 Emissions of Auxiliary Heater (g)']
    if 'series' in str(self._template_path):
      header = [
        'Year', ' Month', ' Day', 'Hour', 'Minute', 'HP Heat Output (kW)',
        'HP Electricity Consumption (kW)', 'HP COP', 'TES Charging Rate (kg/s)',
        'TES Discharging Rate (kg/s)', 'TES Node 1 Temperature', 'TES Node 2 Temperature',
        'TES Node 3 Temperature', 'TES Node 4 Temperature', 'TES Energy Content (J)',
        'TES Energy Content (kWh)', 'TES Energy Content Variation (kWh)',
        'Auxiliary Heater Fuel Flow Rate (kg/s)', 'Auxiliary Heater Energy Input (kW)',
        'HP Operational Cost (CAD)', 'Auxiliary Heater Operational Cost (CAD)',
        'Operational CO2 Emissions of HP (g)',
        'Operational CO2 Emissions of Auxiliary Heater (g)',
        'Return Temperature', 'Demand (kW)']
    header_data = {
      self._input_data['fileOut1']: header,
      self._input_data['fileOut2']: ['Day', 'Operational Daily Emissions from Heat Pumps (g)',
                                     'Operational Daily Emissions from Auxiliary Heater (g)'],
      self._input_data['fileOut3']: ['Month', 'Monthly Operational Costs of Heat Pumps (CAD)',
                                     'Monthly Operational Costs of Auxiliary Heater (CAD)'],
      self._input_data['fileOut4']: ['Month', 'Monthly Fuel Consumption of Auxiliary Heater (m3)'],
      self._input_data['fileOut5']: ['Month', 'Operational Monthly Emissions from Heat Pumps (g)',
                                     'Operational Monthly Emissions from Auxiliary Heater (g)'],
      self._input_data['fileOut6']: ['Day', 'Daily HP Electricity Demand (kWh)'],
      self._input_data['fileOut7']: ['Day', 'Daily Operational Costs of Heat Pumps (CAD)',
                                     'Daily Operational Costs of Auxiliary Heater (CAD)'],
      self._input_data['fileOut8']: ['Month', 'Monthly HP Electricity Demand (kWh)'],
      self._input_data['fileOut9']: ['Day', 'Daily Fuel Consumption of Auxiliary Heater (m3)'],
      self._input_data['fileOut10']: ['Year', 'Month', 'Day', 'Hour', 'HP Electricity Demand (kWh)']
    }

    for file_path, header in header_data.items():
      file_path = file_path.strip("'")
      df = pd.read_csv(file_path, header=None, sep='\s+')
      # ignore ambient temperature for air source series run
      if df.shape[1] > 25 and 'series' in str(self._template_path):
        df.drop(columns=df.columns[-1],
                axis=1,
                inplace=True)
      df.to_csv(file_path, header=header)

  def _update_input_data_with_files(self):
    """
    Updates input data for insel with some files that will
    be written to after insel runs. Also specifies and input file
    which is the Heating Demand (demand.txt) file
    :return:
    """
    self._input_data["HeatingDemand"] = f"'{str(self._demand_path)}'"
    self._input_data["fileOut1"] = f"'{str((self._base_path / 'heat_pumps/technical_performance.csv'))}'"
    self._input_data["fileOut2"] = f"'{str((self._base_path / 'heat_pumps/system_daily_emissions.csv'))}'"
    self._input_data["fileOut3"] = f"'{str((self._base_path / 'heat_pumps/monthly_operational_costs.csv'))}'"
    self._input_data["fileOut4"] = f"'{str((self._base_path / 'heat_pumps/monthly_fossil_fuel_consumptions.csv'))}'"
    self._input_data["fileOut5"] = f"'{str((self._base_path / 'heat_pumps/system_monthly_emissions.csv'))}'"
    self._input_data["fileOut6"] = f"'{str((self._base_path / 'heat_pumps/daily_hp_electricity_demand.csv'))}'"
    self._input_data["fileOut7"] = f"'{str((self._base_path / 'heat_pumps/daily_operational_costs.csv'))}'"
    self._input_data["fileOut8"] = f"'{str((self._base_path / 'heat_pumps/monthly_hp_electricity_demand.csv'))}'"
    self._input_data["fileOut9"] = f"'{str((self._base_path / 'heat_pumps/daily_fossil_fuel_consumption.csv'))}'"
    self._input_data["fileOut10"] = f"'{str((self._base_path / 'heat_pumps/hp_hourly_electricity_demand.csv'))}'"
    # include water temperature for water to water heat pump
    if self._water_temp is not None:
      self._input_data['WaterTemperature'] = f"'{str(self._water_temp)}'"

  def _delete_existing_output_files(self):
    """
    Remove existing out files generated by insel before
    running insel
    :return:
    """
    for key, file_path in self._input_data.items():
      if 'fileOut' in key:
        file_path = file_path.strip("'")
        try:
          os.remove(file_path)
        except OSError:
          pass

  def _compute_max_demand(self):
    """
    Retrieves the maximum demand value from
    the demands text file
    :return: float
    """
    max_demand = -1
    with open(self._demand_path) as file_handler:
      for demand in file_handler.readlines():
        if float(demand) > max_demand:
          max_demand = float(demand)
    return max_demand

  def _update_input_data_with_constants(self):
    with open(self._constants_path) as file:
      constants_dict = yaml.load(file, Loader=yaml.FullLoader)
      for key, value in constants_dict.items():
        if key in ['LowestPossibleLoadFlow', 'HighestPossibleLoadFlow'] and self._water_temp is None:
          continue
        self._input_data[key] = value
    # compute water to water HP specific values
    if 55 <= self._input_data['HPSupTemp'] <= 60:
      self._input_data["HPDeactivationTemperature"] = self._input_data["HPSupTemp"] - 5
      self._input_data["HPReactivationTemperature"] = self._input_data["HPSupTemp"] - 18
    elif 50 <= self._input_data["HPSupTemp"] < 55:
      self._input_data["HPDeactivationTemperature"] = self._input_data["HPSupTemp"] - 5
      self._input_data["HPReactivationTemperature"] = self._input_data["HPSupTemp"] - 13
    elif 45 <= self._input_data["HPSupTemp"] < 50:
      self._input_data["HPDeactivationTemperature"] = self._input_data["HPSupTemp"] - 3
      self._input_data["HPReactivationTemperature"] = self._input_data["HPSupTemp"] - 8
    elif 35 <= self._input_data["HPSupTemp"] < 40:
      self._input_data["HPDeactivationTemperature"] = self._input_data["HPSupTemp"] - 2
      self._input_data["HPReactivationTemperature"] = self._input_data["HPSupTemp"] - 4

    # compute maximum demand. todo: This should come from catalog in the future
    max_demand = self._compute_max_demand()
    # compute TESCapacity
    self._input_data["TESCapacity"] = self._input_data["HoursOfStorageAtMaxDemand"] * (max_demand * 3.6) / (
        (self._input_data["Cp"] / 1000) * self._input_data["TemperatureDifference"])

  def _update_input_data_with_coff(self, a_coefficient: List):
    """
    Updates the user data with coefficients derived from imports
    :param a_coefficient: insel a coefficient values
    Meaning of a coefficient is in the models for air source heat pump
    and water to water source heat pump
    :return:
    """

    self._input_data["a1"] = a_coefficient[0]
    self._input_data["a2"] = a_coefficient[1]
    self._input_data["a3"] = a_coefficient[2]
    self._input_data["a4"] = a_coefficient[3]
    self._input_data["a5"] = a_coefficient[4]
    self._input_data["a6"] = a_coefficient[5]

    # additional coefficients for water to water source
    if self._water_temp is not None:
      self._input_data["a7"] = a_coefficient[6]
      self._input_data["a8"] = a_coefficient[7]
      self._input_data["a9"] = a_coefficient[8]
      self._input_data["a10"] = a_coefficient[9]
      self._input_data["a11"] = a_coefficient[10]

  def _get_user_out_put(self) -> Union[Dict, None]:
    """
    Extracts monthly electricity demand and fossil fuel consumption
    from output files generated by insel
    :return: Dict for json output
    """

    monthly_electricity_df = pd.read_csv(self._input_data['fileOut8'].strip("'")).iloc[:, 2]
    monthly_fossil_df = pd.read_csv(self._input_data['fileOut4'].strip("'")).iloc[:, 2]

    if self._output_path is None:
      return {
        'hourly_electricity_demand': pd.read_csv(self._input_data['fileOut10'].strip("'")).iloc[:, 5].tolist(),
        'monthly_electricity_demand': monthly_electricity_df.tolist(),
        'daily_electricity_demand': pd.read_csv(self._input_data['fileOut6'].strip("'")).iloc[:, 2].tolist(),
        'daily_fossil_consumption': pd.read_csv(self._input_data['fileOut9'].strip("'")).iloc[:, 2].tolist(),
        'monthly_fossil_consumption': monthly_fossil_df.tolist()
      }

    data = [monthly_electricity_df, monthly_fossil_df]
    df = pd.concat(data, axis=1)
    df = pd.concat([df, df.agg(['sum'])])
    s = pd.Series(["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sept", "Oct", "Nov", "Dec", "Total"])
    df = df.set_index([s])
    df.to_csv(self._output_path)



"""
Construction thermal parameters
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.city_model_structure.building_demand.layer import Layer


class Construction:
  """
  Construction class
  """
  def __init__(self):
    self._type = None
    self._layers = None
    self._window_ratio = None
    self._window_frame_ratio = None
    self._window_g_value = None
    self._window_overall_u_value = None
    self._window_type = None

  @property
  def type(self):
    """
    Get construction type
    :return: str
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set construction type
    :param value: str
    """
    self._type = value

  @property
  def layers(self) -> [Layer]:
    """
    Get layers
    :return: [layer]
    """
    return self._layers

  @layers.setter
  def layers(self, value):
    """
    Set layers
    :param value: [layer]
    """
    self._layers = value

  @property
  def window_ratio(self):
    """
    Get window ratio
    :return: dict
    """
    return self._window_ratio

  @window_ratio.setter
  def window_ratio(self, value):
    """
    Set window ratio
    :param value: dict
    """
    self._window_ratio = value

  @property
  def window_frame_ratio(self):
    """
    Get window frame ratio
    :return: float
    """
    return self._window_frame_ratio

  @window_frame_ratio.setter
  def window_frame_ratio(self, value):
    """
    Set window frame ratio
    :param value: float
    """
    self._window_frame_ratio = value

  @property
  def window_g_value(self):
    """
    Get transparent surface g-value
    :return: float
    """
    return self._window_g_value

  @window_g_value.setter
  def window_g_value(self, value):
    """
    Set transparent surface g-value
    :param value: float
    """
    self._window_g_value = value

  @property
  def window_overall_u_value(self):
    """
    Get transparent surface overall U-value in W/m2K
    :return: float
    """
    return self._window_overall_u_value

  @window_overall_u_value.setter
  def window_overall_u_value(self, value):
    """
    Set transparent surface overall U-value in W/m2K
    :param value: float
    """
    self._window_overall_u_value = value

  @property
  def window_type(self):
    """
    Get transparent surface type, 'window' or 'skylight'
    :return: str
    """
    return self._window_type

  @window_type.setter
  def window_type(self, value):
    """
    Set transparent surface type, 'window' or 'skylight'
    :return: str
    """
    self._window_type = value



"""
Time series module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List
from hub.city_model_structure.attributes.record import Record


class TimeSeries:
  """
  TimeSeries class
  """

  def __init__(self, time_series_type=None, records=None):
    self._time_series_type = time_series_type
    self._records = records

  @property
  def time_series_type(self):
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._time_series_type

  @property
  def records(self) -> List[Record]:
    """
    Add explanation here
    :return: List[Record]
    """
    return self._records
